"""Defines the Data Classes used."""

import math
from dataclasses import dataclass
from datetime import datetime
from pprint import pprint as pp
from typing import TypedDict

from typeguard import typechecked

from pyastroweatherio.const import (
    CONDITION,
    CONDITION_PLAIN,
    DEEP_SKY_THRESHOLD,
    LIFTED_INDEX_PLAIN,
    LIFTED_INDEX_RANGE,
    LIFTED_INDEX_VALUE,
    MAG_DEGRATION_MAX,
    SEEING_MAX,
    WIND10M_DIRECTON,
    WIND10M_MAX,
    WIND10M_PLAIN,
    WIND10M_RANGE,
    WIND10M_VALUE,
)


@dataclass
class TimeDataModel(TypedDict):
    """Model for time data"""

    forecast_time: datetime


@typechecked
class TimeData:
    """A representation of the time data of forecasts."""

    def __init__(self, *, data: TimeDataModel):
        self.forecast_time = data["forecast_time"]


class GeoLocationDataModel(TypedDict):
    """Model for the location"""

    latitude: float
    longitude: float
    elevation: float
    timezone_info: str


@typechecked
class GeoLocationData:
    """A representation of the geographic location."""

    def __init__(self, *, data: GeoLocationDataModel):
        self.latitude = data["latitude"]
        self.longitude = data["longitude"]
        self.elevation = data["elevation"]
        self.timezone_info = data["timezone_info"]


class SunDataModel(TypedDict):
    """Model for Sun data"""

    altitude: float
    azimuth: float
    next_rising_astro: datetime
    next_rising_civil: datetime
    next_rising_nautical: datetime
    next_setting_astro: datetime
    next_setting_civil: datetime
    next_setting_nautical: datetime
    previous_rising_astro: datetime
    previous_setting_astro: datetime
    constellation: str


@typechecked
class SunData:
    """A representation of Sun data class."""

    def __init__(self, *, data: SunDataModel):
        self.altitude = data["altitude"]
        self.azimuth = data["azimuth"]
        self.next_rising_astro = data["next_rising_astro"]
        self.next_rising_civil = data["next_rising_civil"]
        self.next_rising_nautical = data["next_rising_nautical"]
        self.next_setting_astro = data["next_setting_astro"]
        self.next_setting_civil = data["next_setting_civil"]
        self.next_setting_nautical = data["next_setting_nautical"]
        self.previous_rising_astro = data["previous_rising_astro"]
        self.previous_setting_astro = data["previous_setting_astro"]
        self.constellation = data["constellation"]


class MoonDataModel(TypedDict):
    """Model for Moon data"""

    altitude: float
    angular_size: float
    avg_angular_size: float
    avg_distance_km: float
    azimuth: float
    distance: float
    distance_km: float
    next_full_moon: datetime
    next_new_moon: datetime
    previous_new_moon: datetime
    next_rising: datetime
    next_setting: datetime
    phase: float
    icon: str
    previous_rising: datetime
    previous_setting: datetime
    relative_distance: float
    relative_size: float
    constellation: str
    next_dark_night: datetime | None


@typechecked
class MoonData:
    """A representation of Moon data class."""

    def __init__(self, *, data: MoonDataModel):
        self.altitude = data["altitude"]
        self.angular_size = data["angular_size"]
        self.avg_angular_size = data["avg_angular_size"]
        self.avg_distance_km = data["avg_distance_km"]
        self.azimuth = data["azimuth"]
        self.distance = data["distance"]
        self.distance_km = data["distance_km"]
        self.next_full_moon = data["next_full_moon"]
        self.next_new_moon = data["next_new_moon"]
        self.previous_new_moon = data["previous_new_moon"]
        self.next_rising = data["next_rising"]
        self.next_setting = data["next_setting"]
        self.phase = data["phase"]
        self.icon = data["icon"]
        self.previous_rising = data["previous_rising"]
        self.previous_setting = data["previous_setting"]
        self.relative_distance = data["relative_distance"]
        self.relative_size = data["relative_size"]
        self.constellation = data["constellation"]
        self.next_dark_night = data["next_dark_night"]


class DarknessDataModel(TypedDict):
    """Model for deep sky darkness"""

    deep_sky_darkness_moon_rises: bool
    deep_sky_darkness_moon_sets: bool
    deep_sky_darkness_moon_always_up: bool
    deep_sky_darkness_moon_always_down: bool
    deep_sky_darkness: float


@typechecked
class DarknessData:
    """A representation of darkness data class."""

    def __init__(self, *, data: DarknessDataModel):
        self.deep_sky_darkness_moon_rises = data["deep_sky_darkness_moon_rises"]
        self.deep_sky_darkness_moon_sets = data["deep_sky_darkness_moon_sets"]
        self.deep_sky_darkness_moon_always_up = data["deep_sky_darkness_moon_always_up"]
        self.deep_sky_darkness_moon_always_down = data["deep_sky_darkness_moon_always_down"]
        self.deep_sky_darkness = data["deep_sky_darkness"]


class ConditionDataModel(TypedDict):
    """Model for weather conditions"""

    cloudcover: float
    cloud_area_fraction: float
    cloud_area_fraction_high: float
    cloud_area_fraction_low: float
    cloud_area_fraction_medium: float
    fog_area_fraction: float
    fog2m: float
    seeing: float
    transparency: float
    lifted_index: float
    condition_percentage: int
    rh2m: float
    wind_speed: float
    wind_from_direction: float
    temp2m: float
    dewpoint2m: float
    weather: str
    weather6: str
    precipitation_amount: float
    precipitation_amount6: float


@typechecked
class ConditionData:
    """A representation of the condition base class."""

    def __init__(self, *, data: ConditionDataModel):
        self.cloudcover = data["cloudcover"]
        self.cloud_area_fraction = data["cloud_area_fraction"]
        self.cloud_area_fraction_high = data["cloud_area_fraction_high"]
        self.cloud_area_fraction_low = data["cloud_area_fraction_low"]
        self.cloud_area_fraction_medium = data["cloud_area_fraction_medium"]
        self.fog_area_fraction = data["fog_area_fraction"]
        self.fog2m = data["fog2m"]
        self._seeing = data["seeing"]
        self._transparency = data["transparency"]
        self._lifted_index = data["lifted_index"]
        self.condition_percentage = data["condition_percentage"]
        self.rh2m = data["rh2m"]
        self.wind_speed = data["wind_speed"]
        self.wind_from_direction = data["wind_from_direction"]
        self.temp2m = data["temp2m"]
        self._dewpoint2m = data["dewpoint2m"]
        self._weather = data["weather"]
        self._weather6 = data["weather6"]
        self.precipitation_amount = data["precipitation_amount"]
        self.precipitation_amount6 = data["precipitation_amount6"]

    # #########################################################################
    # Condition
    # #########################################################################
    @property
    def cloudcover_percentage(self) -> int:
        """Return Cloud Cover Percentage."""

        return int(self.cloudcover)

    @property
    def cloudless_percentage(self) -> int:
        """Return Cloudless Percentage."""

        return 100 - int(self.cloudcover)

    @property
    def cloud_area_fraction_percentage(self) -> int:
        """Return Cloud Cover Percentage."""

        return int(self.cloud_area_fraction)

    @property
    def cloud_area_fraction_high_percentage(self) -> int:
        """Return Cloud Cover Percentage."""

        return int(self.cloud_area_fraction_high)

    @property
    def cloud_area_fraction_medium_percentage(self) -> int:
        """Return Cloud Cover Percentage."""

        return int(self.cloud_area_fraction_medium)

    @property
    def cloud_area_fraction_low_percentage(self) -> int:
        """Return Cloud Cover Percentage."""

        return int(self.cloud_area_fraction_low)

    @property
    def fog_area_fraction_percentage(self) -> int:
        """Return Fog Area Percentage."""

        return int(self.fog_area_fraction)

    @property
    def fog2m_area_fraction_percentage(self) -> int:
        """Return Fog Density Percentage."""

        return int(self.fog2m)

    @property
    def seeing(self) -> float:
        """Return Seeing."""

        return round(self._seeing, 2)

    @seeing.setter
    def seeing(self, new_value: float):
        """Never called"""
        self._seeing = new_value

    @property
    def seeing_percentage(self) -> int:
        """Return Seeing Percentage."""

        return int(100 - self._seeing * 100 / SEEING_MAX)

    @property
    def transparency(self) -> float:
        """Return Transparency."""

        return round(self._transparency, 2)

    @transparency.setter
    def transparency(self, new_value: float):
        """Never called"""
        self._transparency = new_value

    @property
    def transparency_percentage(self) -> int:
        """Return Transparency."""

        return int(100 - self._transparency * 100 / MAG_DEGRATION_MAX)

    @property
    def lifted_index(self) -> float:
        """Return Lifted Index."""

        return round(self._lifted_index, 2)

    @lifted_index.setter
    def lifted_index(self, new_value: float):
        """Never called"""
        self._lifted_index = new_value

    @property
    def wind10m_speed(self) -> float:
        """Return 10m Wind Speed."""

        return self.wind_speed

    @property
    def calm_percentage(self) -> int:
        """Return 10m Wind Speed."""

        return int(100 - self.wind_speed * (100 / WIND10M_MAX))

    @property
    def wind10m_direction(self) -> str:
        """Return 10m Wind Direction."""

        direction = self.wind_from_direction
        direction += 22.5
        direction = direction % 360
        direction = int(direction / 45)  # values 0 to 7
        return WIND10M_DIRECTON[max(0, min(7, direction))]

    @property
    def dewpoint2m(self) -> float:
        """Return 2m Dew Point."""

        return round(self._dewpoint2m, 1)

    @dewpoint2m.setter
    def dewpoint2m(self, new_value: float):
        """Never called"""
        self._dewpoint2m = new_value

    @property
    def weather(self) -> str:
        """Return Current Weather."""

        return self._weather.replace("_", " ").capitalize()

    @weather.setter
    def weather(self, new_value: float):
        """Never called"""
        self._weather = new_value

    @property
    def weather6(self) -> str:
        """Return Current Weather."""

        return self._weather6.replace("_", " ").capitalize()

    @weather6.setter
    def weather6(self, new_value: float):
        """Never called"""
        self._weather6 = new_value


class UpTonightDSODataModel(TypedDict):
    """Model for DSO objects"""

    id: str
    target_name: str
    type: str
    constellation: str
    size: float
    visual_magnitude: float
    meridian_transit: datetime | str
    meridian_antitransit: datetime | str
    foto: float


@typechecked
class UpTonightDSOData:
    """A representation of uptonight DSO."""

    def __init__(self, *, data: UpTonightDSODataModel):
        self.id = data["id"]
        self.target_name = data["target_name"]
        self.type = data["type"]
        self.constellation = data["constellation"]
        self.size = data["size"]
        self.visual_magnitude = data["visual_magnitude"]
        self.meridian_transit = data["meridian_transit"]
        self.meridian_antitransit = data["meridian_antitransit"]
        self.foto = data["foto"]


class UpTonightBodiesDataModel(TypedDict):
    """Model for bodies"""

    target_name: str
    max_altitude: float
    azimuth: float
    max_altitude_time: datetime | str
    visual_magnitude: float
    meridian_transit: datetime | str
    foto: float


@typechecked
class UpTonightBodiesData:
    """A representation of uptonight bodies."""

    def __init__(self, *, data: UpTonightBodiesDataModel):
        self.target_name = data["target_name"]
        self.max_altitude = data["max_altitude"]
        self.azimuth = data["azimuth"]
        self.max_altitude_time = data["max_altitude_time"]
        self.visual_magnitude = data["visual_magnitude"]
        self.meridian_transit = data["meridian_transit"]
        self.foto = data["foto"]


class UpTonightCometsDataModel(TypedDict):
    """Model for comets"""

    designation: str
    distance_au_earth: float
    distance_au_sun: float
    absolute_magnitude: float
    visual_magnitude: float
    altitude: float
    azimuth: float
    rise_time: datetime | str
    set_time: datetime | str


@typechecked
class UpTonightCometsData:
    """A representation of uptonight comets."""

    def __init__(self, *, data: UpTonightCometsDataModel):
        self.designation = data["designation"]
        self.distance_au_earth = data["distance_au_earth"]
        self.distance_au_sun = data["distance_au_sun"]
        self.absolute_magnitude = data["absolute_magnitude"]
        self.visual_magnitude = data["visual_magnitude"]
        self.altitude = data["altitude"]
        self.azimuth = data["azimuth"]
        self.rise_time = data["rise_time"]
        self.set_time = data["set_time"]


class LocationDataModel(TypedDict):
    """Model for location conditions data"""

    time_data: TimeData
    time_shift: int
    forecast_length: int
    location_data: GeoLocationData
    sun_data: SunData
    moon_data: MoonData
    darkness_data: DarknessData
    night_duration_astronomical: float
    deepsky_forecast: list
    condition_data: ConditionData
    uptonight: list
    uptonight_bodies: list
    uptonight_comets: list


@typechecked
class LocationData:
    """A representation of the Location AstroWeather Data."""

    def __init__(self, *, data: LocationDataModel):
        self.time_data = data["time_data"]
        self.time_shift = data["time_shift"]
        self.forecast_length = data["forecast_length"]
        self.location_data = data["location_data"]
        self.sun_data = data["sun_data"]
        self.moon_data = data["moon_data"]
        self.darkness_data = data["darkness_data"]
        self.night_duration_astronomical = data["night_duration_astronomical"]
        self.deepsky_forecast = data["deepsky_forecast"]
        self.condition_data = data["condition_data"]
        self._uptonight = data["uptonight"]
        self._uptonight_bodies = data["uptonight_bodies"]
        self._uptonight_comets = data["uptonight_comets"]

    # #########################################################################
    # Time data
    # #########################################################################
    @property
    def forecast_time(self) -> datetime:
        return self.time_data.forecast_time

    # #########################################################################
    # Location
    # #########################################################################
    @property
    def latitude(self) -> float:
        """Return Latitude."""

        return self.location_data.latitude

    @property
    def longitude(self) -> float:
        """Return Longitude."""

        return self.location_data.longitude

    @property
    def elevation(self) -> float:
        """Return Elevation."""

        return self.location_data.elevation

    # #########################################################################
    # Condition
    # #########################################################################
    @property
    def wind10m_speed_plain(self) -> str:
        """Return wind speed plain."""

        wind_speed_value = 0
        for (start, end), derate in zip(WIND10M_RANGE, WIND10M_VALUE):
            if start <= self.condition_data.wind_speed <= end:
                wind_speed_value = derate

        return WIND10M_PLAIN[max(0, min(7, wind_speed_value - 1))]

    @property
    def lifted_index_plain(self) -> str:
        """Return Lifted Index plain."""

        lifted_index_value = 0
        for (start, end), derate in zip(LIFTED_INDEX_RANGE, LIFTED_INDEX_VALUE):
            if start <= self.condition_data.lifted_index <= end:
                lifted_index_value = derate

        return LIFTED_INDEX_PLAIN[max(0, min(7, lifted_index_value - 1))]

    @property
    def deep_sky_view(self) -> bool:
        """Return True if Deep Sky should be possible."""

        if self.condition_percentage >= DEEP_SKY_THRESHOLD:
            return True
        return False

    @property
    def condition_plain(self) -> str:
        """Return Current View Conditions."""

        if self.condition_percentage > 80:
            return CONDITION_PLAIN[0].capitalize()
        if self.condition_percentage > 60:
            return CONDITION_PLAIN[1].capitalize()
        if self.condition_percentage > 40:
            return CONDITION_PLAIN[2].capitalize()
        if self.condition_percentage > 20:
            return CONDITION_PLAIN[3].capitalize()
        return CONDITION_PLAIN[4].capitalize()

    @property
    def condition_percentage(self) -> int:
        return self.condition_data.condition_percentage

    @property
    def cloudcover_percentage(self) -> int:
        return self.condition_data.cloudcover_percentage

    @property
    def cloudless_percentage(self) -> int:
        return self.condition_data.cloudless_percentage

    @property
    def cloud_area_fraction_percentage(self) -> int:
        return self.condition_data.cloud_area_fraction_percentage

    @property
    def cloud_area_fraction_high_percentage(self) -> int:
        return self.condition_data.cloud_area_fraction_high_percentage

    @property
    def cloud_area_fraction_medium_percentage(self) -> int:
        return self.condition_data.cloud_area_fraction_medium_percentage

    @property
    def cloud_area_fraction_low_percentage(self) -> int:
        return self.condition_data.cloud_area_fraction_low_percentage

    @property
    def fog_area_fraction_percentage(self) -> int:
        return self.condition_data.fog_area_fraction_percentage

    @property
    def fog2m_area_fraction_percentage(self) -> int:
        return self.condition_data.fog2m_area_fraction_percentage

    @property
    def seeing(self) -> float:
        return self.condition_data.seeing

    @property
    def seeing_percentage(self) -> int:
        return self.condition_data.seeing_percentage

    @property
    def transparency(self) -> float:
        return self.condition_data.transparency

    @property
    def transparency_percentage(self) -> int:
        return self.condition_data.transparency_percentage

    @property
    def lifted_index(self) -> float:
        return self.condition_data.lifted_index

    @property
    def calm_percentage(self) -> int:
        return self.condition_data.calm_percentage

    @property
    def wind10m_direction(self) -> str:
        return self.condition_data.wind10m_direction

    @property
    def wind10m_speed(self) -> float:
        return self.condition_data.wind10m_speed

    @property
    def temp2m(self) -> float:
        return self.condition_data.temp2m

    @property
    def rh2m(self) -> float:
        return self.condition_data.rh2m

    @property
    def dewpoint2m(self) -> float:
        return self.condition_data.dewpoint2m

    @property
    def weather(self) -> str:
        return self.condition_data.weather

    @property
    def weather6(self) -> str:
        return self.condition_data.weather6

    @property
    def precipitation_amount(self) -> float:
        return self.condition_data.precipitation_amount

    @property
    def precipitation_amount6(self) -> float:
        return self.condition_data.precipitation_amount6

    # #########################################################################
    # Sun
    # #########################################################################
    @property
    def sun_altitude(self) -> float:
        """Return Sun Altitude."""

        return round(self.sun_data.altitude, 3)

    @property
    def sun_azimuth(self) -> float:
        """Return sun Azimuth."""

        return round(self.sun_data.azimuth, 3)

    @property
    def sun_next_rising_astro(self) -> datetime:
        """Return Sun Next Rising Astronomical."""

        return self.sun_data.next_rising_astro

    @property
    def sun_next_rising(self) -> datetime:
        """Return Sun Next Rising Civil."""

        return self.sun_data.next_rising_civil

    @property
    def sun_next_rising_nautical(self) -> datetime:
        """Return Sun Next Rising Nautical."""

        return self.sun_data.next_rising_nautical

    @property
    def sun_next_setting_astro(self) -> datetime:
        """Return Sun Next Setting Astronomical."""

        return self.sun_data.next_setting_astro

    @property
    def sun_next_setting(self) -> datetime:
        """Return Next Setting Civil."""

        return self.sun_data.next_setting_civil

    @property
    def sun_next_setting_nautical(self) -> datetime:
        """Return Sun Next Setting Nautical."""

        return self.sun_data.next_setting_nautical

    @property
    def sun_constellation(self) -> str:
        """Return Sun Constellation."""

        return self.sun_data.constellation

    # #########################################################################
    # Moon
    # #########################################################################
    @property
    def moon_altitude(self) -> float:
        """Return Moon Altitude."""

        return round(self.moon_data.altitude, 3)

    @property
    def moon_angular_size(self) -> float:
        """Return Moon Angular Size in Minutes."""

        return round(self.moon_data.angular_size, 3)

    @property
    def moon_azimuth(self) -> float:
        """Return Moon Azimuth."""

        return round(self.moon_data.azimuth, 3)

    @property
    def moon_distance_km(self) -> float:
        """Return Moon Distance in km."""

        return round(self.moon_data.distance_km, 0)

    @property
    def moon_next_full_moon(self) -> datetime:
        """Return Moon Next Full Moon."""

        return self.moon_data.next_full_moon

    @property
    def moon_next_new_moon(self) -> datetime:
        """Return Moon Next New Moon."""

        return self.moon_data.next_new_moon

    @property
    def moon_next_rising(self) -> datetime:
        """Return Moon Next Rising."""

        return self.moon_data.next_rising

    @property
    def moon_next_setting(self) -> datetime:
        """Return Moon Next Setting."""

        return self.moon_data.next_setting

    @property
    def moon_phase(self) -> float:
        """Return Moon Phase."""

        return round(self.moon_data.phase, 1)

    @property
    def moon_icon(self) -> str:
        """Return Moon Phase."""

        return self.moon_data.icon

    @property
    def moon_relative_size(self) -> float:
        """Return Moon Relative Size in %."""

        return round(self.moon_data.relative_size * 100 - 100, 3)

    @property
    def moon_relative_distance(self) -> float:
        """Return Moon Relative Distance in %."""

        return round(self.moon_data.relative_distance * 100 - 100, 3)

    @property
    def moon_constellation(self) -> str:
        """Return Moon Constellation."""

        return self.moon_data.constellation

    @property
    def moon_next_dark_night(self) -> datetime:
        """Returns the next dark night."""

        return self.moon_data.next_dark_night

    # #########################################################################
    # Darkness
    # #########################################################################
    @property
    def deep_sky_darkness_moon_rises(self) -> bool:
        """Returns true if moon rises during astronomical night."""

        return self.darkness_data.deep_sky_darkness_moon_rises

    @property
    def deep_sky_darkness_moon_sets(self) -> bool:
        """Returns true if moon sets during astronomical night."""

        return self.darkness_data.deep_sky_darkness_moon_sets

    @property
    def deep_sky_darkness_moon_always_up(self) -> bool:
        """Returns true if moon is up during astronomical night."""

        return self.darkness_data.deep_sky_darkness_moon_always_up

    @property
    def deep_sky_darkness_moon_always_down(self) -> bool:
        """Returns true if moon is down during astronomical night."""

        return self.darkness_data.deep_sky_darkness_moon_always_down

    @property
    def deep_sky_darkness(self) -> float:
        """Returns the remaining timespan of deep sky darkness."""

        return self.darkness_data.deep_sky_darkness

    # #########################################################################
    # Deep Sky Forecast
    # #########################################################################
    @property
    def deepsky_forecast_today(self) -> int:
        """Return Forecas Today in Percent."""

        nightly_condition_sum = 0
        if len(self.deepsky_forecast) > 0:
            for nightly_condition in self.deepsky_forecast[0].nightly_conditions:
                nightly_condition_sum += nightly_condition
            return int(round(nightly_condition_sum / len(self.deepsky_forecast[0].nightly_conditions)))

    @property
    def deepsky_forecast_today_dayname(self):
        """Return Forecast Todays Dayname."""

        if len(self.deepsky_forecast) > 0:
            nightly_conditions = self.deepsky_forecast[0]
            return nightly_conditions.dayname

    @property
    def deepsky_forecast_today_plain(self):
        """Return Forecast Today."""

        out = ""
        if len(self.deepsky_forecast) > 0:
            for nightly_condition in self.deepsky_forecast[0].nightly_conditions:
                out += CONDITION[4 - math.floor(nightly_condition / 20)].capitalize()
        return out

    @property
    def deepsky_forecast_today_desc(self):
        """Return Forecast Today Description."""

        if len(self.deepsky_forecast) > 0:
            nightly_conditions = self.deepsky_forecast[0]
            return nightly_conditions.weather.replace("_", " ").capitalize()

    @property
    def deepsky_forecast_today_precipitation_amount6(self) -> float:
        """Return Forecast Today Precipitation Amount6."""

        if len(self.deepsky_forecast) > 0:
            nightly_conditions = self.deepsky_forecast[0]
            return nightly_conditions.precipitation_amount6

    @property
    def deepsky_forecast_tomorrow(self) -> int:
        """Return Forecas Tomorrow in Percentt."""

        nightly_condition_sum = 0
        if len(self.deepsky_forecast) > 1:
            for nightly_condition in self.deepsky_forecast[1].nightly_conditions:
                nightly_condition_sum += nightly_condition
            return int(round(nightly_condition_sum / len(self.deepsky_forecast[1].nightly_conditions)))

    @property
    def deepsky_forecast_tomorrow_dayname(self):
        """Return Forecast Todays Dayname."""

        if len(self.deepsky_forecast) > 1:
            nightly_conditions = self.deepsky_forecast[1]
            return nightly_conditions.dayname

    @property
    def deepsky_forecast_tomorrow_plain(self):
        """Return Forecast Tomorrow."""

        out = ""
        if len(self.deepsky_forecast) > 1:
            for nightly_condition in self.deepsky_forecast[1].nightly_conditions:
                out += CONDITION[4 - math.floor(nightly_condition / 20)].capitalize()
        return out

    @property
    def deepsky_forecast_tomorrow_desc(self):
        """Return Forecast Tomorrow Description."""

        if len(self.deepsky_forecast) > 1:
            nightly_conditions = self.deepsky_forecast[1]
            return nightly_conditions.weather.replace("_", " ").capitalize()

    @property
    def deepsky_forecast_tomorrow_precipitation_amount6(self) -> float:
        """Return Forecast Today Precipitation Amount6."""

        if len(self.deepsky_forecast) > 0:
            nightly_conditions = self.deepsky_forecast[1]
            return nightly_conditions.precipitation_amount6

    # #########################################################################
    # UpTonight
    # #########################################################################
    @property
    def uptonight(self) -> int:
        """Return the number of best DSOs for tonight."""

        return len(self._uptonight)

    @property
    def uptonight_list(self) -> list[UpTonightDSOData]:
        """Return the list of UpTonight targets."""

        return self._uptonight

    # #########################################################################
    # UpTonight Bodies
    # #########################################################################
    @property
    def uptonight_bodies(self) -> int:
        """Return the number of best BODIEs for tonight."""

        return len(self._uptonight_bodies)

    @property
    def uptonight_bodies_list(self) -> list[UpTonightBodiesData]:
        """Return the list of UpTonight bodies."""

        return self._uptonight_bodies

    # #########################################################################
    # UpTonight Comets
    # #########################################################################
    @property
    def uptonight_comets(self) -> int:
        """Return the number of best comets for tonight."""

        return len(self._uptonight_comets)

    @property
    def uptonight_comets_list(self) -> list[UpTonightCometsData]:
        """Return the list of UpTonight comets."""

        return self._uptonight_comets


class ForecastDataModel(TypedDict):
    """Model for forecast data"""

    time_data: TimeData
    hour: int
    condition_data: ConditionData


@typechecked
class ForecastData:
    """A representation of 3-Hour Based Forecast AstroWeather Data."""

    def __init__(self, *, data: ForecastDataModel):
        self.time_data = data["time_data"]
        self.hour = data["hour"]
        self.condition_data = data["condition_data"]

    # #########################################################################
    # Time data
    # #########################################################################
    @property
    def forecast_time(self) -> datetime:
        return self.time_data.forecast_time

    # #########################################################################
    # Forecast data
    # #########################################################################
    @property
    def deep_sky_view(self) -> bool:
        """Return True if Deep Sky should be possible."""

        if self.condition_percentage <= DEEP_SKY_THRESHOLD:
            return True
        return False

    @property
    def condition_percentage(self) -> int:
        return self.condition_data.condition_percentage

    @property
    def cloudcover_percentage(self) -> int:
        return self.condition_data.cloudcover_percentage

    @property
    def cloudless_percentage(self) -> int:
        return self.condition_data.cloudless_percentage

    @property
    def cloud_area_fraction_percentage(self) -> int:
        return self.condition_data.cloud_area_fraction_percentage

    @property
    def cloud_area_fraction_high_percentage(self) -> int:
        return self.condition_data.cloud_area_fraction_high_percentage

    @property
    def cloud_area_fraction_medium_percentage(self) -> int:
        return self.condition_data.cloud_area_fraction_medium_percentage

    @property
    def cloud_area_fraction_low_percentage(self) -> int:
        return self.condition_data.cloud_area_fraction_low_percentage

    @property
    def fog_area_fraction_percentage(self) -> int:
        return self.condition_data.fog_area_fraction_percentage

    @property
    def fog2m_area_fraction_percentage(self) -> int:
        return self.condition_data.fog2m_area_fraction_percentage

    @property
    def seeing(self) -> float:
        return self.condition_data.seeing

    @property
    def seeing_percentage(self) -> int:
        return self.condition_data.seeing_percentage

    @property
    def transparency(self) -> float:
        return self.condition_data.transparency

    @property
    def transparency_percentage(self) -> int:
        return self.condition_data.transparency_percentage

    @property
    def lifted_index(self) -> float:
        return self.condition_data.lifted_index

    @property
    def calm_percentage(self) -> int:
        return self.condition_data.calm_percentage

    @property
    def wind10m_direction(self) -> str:
        return self.condition_data.wind10m_direction

    @property
    def wind10m_speed(self) -> float:
        return self.condition_data.wind10m_speed

    @property
    def temp2m(self) -> float:
        return self.condition_data.temp2m

    @property
    def rh2m(self) -> float:
        return self.condition_data.rh2m

    @property
    def dewpoint2m(self) -> float:
        return self.condition_data.dewpoint2m

    @property
    def weather(self) -> str:
        return self.condition_data.weather

    @property
    def weather6(self) -> str:
        return self.condition_data.weather6

    @property
    def precipitation_amount(self) -> float:
        return self.condition_data.precipitation_amount

    @property
    def precipitation_amount6(self) -> float:
        return self.condition_data.precipitation_amount6


class NightlyConditionsDataModel(TypedDict):
    """Model for nightly conditions data"""

    dayname: str
    hour: int
    nightly_conditions: list
    weather: str
    precipitation_amount6: float


@typechecked
class NightlyConditionsData:
    """A representation of nights Sky Quality Data."""

    def __init__(self, *, data: NightlyConditionsDataModel):
        self.dayname = data["dayname"]
        self.hour = data["hour"]
        self.nightly_conditions = data["nightly_conditions"]
        self._weather = data["weather"]
        self.precipitation_amount6 = data["precipitation_amount6"]

    @property
    def weather(self) -> str:
        """Return Current Weather."""
        return self._weather.replace("_", " ").capitalize()

    @weather.setter
    def weather(self, new_value: str):
        """Never called"""
        self._weather = new_value



"""Contains Helper functions for AstroWeather."""

import bisect
import logging
import math
from datetime import UTC, datetime, timedelta
from math import degrees as deg

import ephem
from ephem import degree
from typeguard import typechecked
from zoneinfo import ZoneInfo

from pyastroweatherio.const import (
    ASTRONOMICAL_DUSK_DAWN,
    CIVIL_DUSK_DAWN,
    DARK_NIGHT_MAX_MOON_ALT,
    DARK_NIGHT_MAX_MOON_PHASE,
    MAG_DEGRATION_MAX,
    NAUTICAL_DUSK_DAWN,
    SEEING_MAX,
)
from pyastroweatherio.dataclasses import (
    DarknessData,
    DarknessDataModel,
    MoonData,
    MoonDataModel,
    SunData,
    SunDataModel,
)

_LOGGER = logging.getLogger(__name__)


class ConversionFunctions:
    """Convert between different units."""

    async def epoch_to_datetime(self, value) -> str:
        """Converts EPOC time to Date Time String."""

        return datetime.datetime.fromtimestamp(int(value)).strftime("%Y-%m-%d %H:%M:%S")


class AtmosphericRoutines:
    """Calculate atmospheric attributes."""

    def __init__(
        self,
    ):
        _LOGGER.debug("AtmosphericRoutines calculation mode active")

        # Relative humidity (0.0 to 1.0)
        # Temperature in Celsius
        # Cloud cover fraction (0.0 to 1.0)
        # Wind speed in meters per second
        # Humidity in percent
        # Altitude in meters
        # Dew point temperature in Celsius
        # Air pressure at sea level in hPa
        # aerosol_density in kg/m^3

    # #####################################################
    # Calculate lifted index
    # #####################################################
    @typechecked
    async def calculate_lifted_index(
        self, temperature, altitude, dew_point_temperature, air_pressure_at_sea_level
    ) -> None | float:
        """Calculate atmospheric lifted index."""
        # https://en.wikipedia.org/wiki/Lifted_index

        values = {
            "temperature": temperature,
            "altitude": altitude,
            "dew_point_temperature": dew_point_temperature,
            "air_pressure_at_sea_level": air_pressure_at_sea_level,
        }

        missing = [name for name, val in values.items() if val is None or (isinstance(val, float) and math.isnan(val))]

        if missing:
            _LOGGER.warning(f"calculate_lifted_index: The following variables are None or NaN: {', '.join(missing)}")
            return None
            
        # Constants
        env_temp_500mb = -20  # Celsius (environmental temperature at 500 mb level)

        # Calculate saturation vapor pressure at surface temperature
        # Checked with https://www.weather.gov/epz/wxcalc_vaporpressure
        # es = self._calculate_vapor_pressure(temperature)

        # Calculate actual Vapor Pressure at surface
        # Checked with https://www.weather.gov/epz/wxcalc_vaporpressure
        e = self._calculate_vapor_pressure(dew_point_temperature)  # 6.112 * (10 ** (7.5 * (Td - Tn) / (Td - 35.85)))

        # Calculate Mixing Ratio at Surface in grams per kilogram
        # Checked with https://www.weather.gov/epz/wxcalc_mixingratio
        w = self._calculate_mixing_ratio(e, air_pressure_at_sea_level)

        # Calculate Lifting Condensation Level
        lcl = self._calculate_lifting_condensation_level(w, air_pressure_at_sea_level)

        # Calculate temperature of lifted parcel at 500 mb level
        lifted_temp_500mb = (
            env_temp_500mb + (temperature - lcl) * 0.5
        )  # Assumption: 500 mb is halfway through the troposphere

        # Calculate Lifted Index
        lifted_index = (env_temp_500mb - lifted_temp_500mb) * 2 + 7

        # Ensure lifted index is within the valid range [-7, 7]
        lifted_index = max(-7, min(7, lifted_index))

        # _LOGGER.debug(f"Vapor Pressure: {e}")
        # _LOGGER.debug(f"Mixing Ratio: {w}")
        # _LOGGER.debug(f"Lifting Condensation Level: {lcl}")
        # _LOGGER.debug(f"Temperature of Lifted Parcel: {lifted_temp_500mb}")
        # _LOGGER.debug(f"Lifted Index: {lifted_index}")

        return lifted_index

    # #####################################################
    # Calculate magniture degradation based on transparency
    # #####################################################
    @typechecked
    async def magnitude_degradation(
        self,
        temperature,
        humidity,
        cloud_cover,
        wind_speed,
        altitude,
        dew_point_temperature,
        air_pressure_at_sea_level,
    ) -> None | float:
        """
        Calculates the magnitude_degradation of the atmosphere.
        This algorithm first calculates the lifted index and the seeing and uses them to calculate
        the transparency from which the magnitude degradation is derived from.

        Args:
        - temperature: Surface temperature in Celsius.
        - humidity: Humidity in Percent.
        - dew_point_temperature: Dew point temperature in Celsius.
        - wind_speed: Wind speed in meters per second.
        - altitude: Altitude in meters.
        - air_pressure_at_sea_level: Air pressure at sea level.
        - cloud_cover: Cloud cover in Percent.
        - lifted_index:
        - seeing:

        Returns:
        - magnitude_degradation: In magnitude
        """

        values = {
            "temperature": temperature,
            "humidity": humidity,
            "cloud_cover": cloud_cover,
            "wind_speed": wind_speed,
            "altitude": altitude,
            "dew_point_temperature": dew_point_temperature,
            "air_pressure_at_sea_level": air_pressure_at_sea_level,
        }

        missing = [name for name, val in values.items() if val is None or (isinstance(val, float) and math.isnan(val))]

        if missing:
            _LOGGER.warning(f"magnitude_degradation: The following variables are None or NaN: {', '.join(missing)}")
            return None
        
        lifted_index = await self.calculate_lifted_index(
            temperature, altitude, dew_point_temperature, air_pressure_at_sea_level
        )
        seeing = await self.calculate_seeing(
            temperature,
            humidity,
            dew_point_temperature,
            wind_speed,
            cloud_cover,
            altitude,
            air_pressure_at_sea_level,
        )

        # Calculate transparency
        transparency = self._calculate_transparency(
            humidity,
            temperature,
            cloud_cover,
            wind_speed,
            altitude,
            dew_point_temperature,
            air_pressure_at_sea_level,
            lifted_index,
            seeing,
        )

        # Convert transparency to magnitude degradation
        magnitude_degradation = self._transparency_to_magnitude_degradation(transparency)

        # _LOGGER.debug(
        #     "Magnitude Degradation: {:.2f} mag (".format(magnitude_degradation)
        #     + "Lifted Index (LI): {:.2f} °C, ".format(lifted_index)
        #     + "Seeing: {:.2f} arcsec, ".format(seeing)
        #     + "Estimated Atmospheric Transparency: {:.2f})".format(transparency)
        # )

        return magnitude_degradation

    # #####################################################
    # Calculate atmospheric seeing
    # #####################################################
    @typechecked
    async def calculate_seeing(
        self,
        temperature,
        humidity,
        dew_point_temperature,
        wind_speed,
        cloud_cover,
        altitude,
        air_pressure_at_sea_level,
    ) -> None | float:
        """
        Calculated seeing of the atmosphere. This algorithm first calculates the seeing factor based on temperature,
        humidity, wind speed and altitude above sea level. The seeing factor is then used to calculate the astronomical
        seeing in arcseconds. The empirical relationship used here states that the seeing in arcseconds is approximately
        equal to the reciprocal of the seeing factor multiplied by a conversion factor of 0.98.

        Used by: magnitude degradation

        Args:
        - temperature: Surface temperature in Celsius.
        - humidity: Humidity in Percent.
        - dew_point_temperature: Dew point temperature in Celsius.
        - wind_speed: Wind speed in meters per second.
        - cloud_cover: Cloud cover in Percent.
        - altitude: Altitude in meters.
        - air_pressure_at_sea_level: Air pressure at sea level.

        Returns:
        - seeing: In arcsecs

        Flow:
        - _calculate_water_vapor_pressure
        - adjusted_pressure
        - relative_pressure
        - seeing_factor
        - seeing
        """
        
        values = {
            "temperature": temperature,
            "humidity": humidity,
            "dew_point_temperature": dew_point_temperature,
            "wind_speed": wind_speed,
            "cloud_cover": cloud_cover,
            "altitude": altitude,
            "air_pressure_at_sea_level": air_pressure_at_sea_level,
        }

        missing = [name for name, val in values.items() if val is None or (isinstance(val, float) and math.isnan(val))]

        if missing:
            _LOGGER.warning(f"calculate_seeing: The following variables are None or NaN: {', '.join(missing)}")
            return None
        
        # Constants
        C = 6.5  # 1.7

        water_vapor_pressure = self._calculate_water_vapor_pressure(dew_point_temperature, humidity)

        # adjusted_pressure = air_pressure_at_sea_level * math.exp(-0.00012 * altitude)
        adjusted_pressure = self._calculate_adjusted_pressure(air_pressure_at_sea_level, altitude)
        relative_pressure = adjusted_pressure / air_pressure_at_sea_level

        if wind_speed == 0:
            wind_speed = 0.01
        seeing_factor = C * (water_vapor_pressure / 10) ** 0.25 * (wind_speed / 10) ** 0.75 * relative_pressure
        seeing = 0.98 / seeing_factor

        # _LOGGER.debug(
        #     "Seeing: {:.2f} arcsec (".format(seeing)
        #     + "Water Vapor Pressure: {:.2f} mbar, ".format(water_vapor_pressure)
        #     + "Wind Speed: {:.2f} m/s, ".format(wind_speed)
        #     + "Relative Pressure: {:.2f} mbar, ".format(relative_pressure)
        #     + "Seeing Factor: {:.2f})".format(seeing_factor)
        # )

        if seeing > SEEING_MAX:
            seeing = SEEING_MAX  # max out seeing

        return seeing

    # #####################################################
    # Calculate fog density at 2m
    # #####################################################
    @typechecked
    async def calculate_fog_density(self, temp2m, rh2m, dewpoint2m, wind_speed) -> float:
        """
        To calculate fog density, one can use a method based on the Liquid Water Content (LWC)
        in fog, which depends on the difference between air temperature and dew point, relative
        humidity, and wind speed. While there is no universally agreed-upon formula for fog
        density, the implemented approximate method can provide a rough measure.
        Note that actual fog density can be influenced by more complex microclimatic factors,
        so this approach serves as a simplified approximation.

        Args:
        - temp2m: Surface temperature in Celsius.
        - rh2m: Humidity in Percent.
        - dewpoint2m: Dew point temperature in Celsius.
        - wind_speed: Wind speed in meters per second.

        Returns:
        - adjusted_fog_density
        """

        # Constants
        # A normalizing constant chosen empirically to scale the fog density to a reasonable
        # range. Adjust this based on the desired scale of output (e.g., 0.1 for very light
        # fog, 1 for dense fog).
        k = 0.50

        # Controls how sensitive the density calculation is to the temperature-dew point
        # difference. Smaller values of a make fog density increase faster as the temperature
        # approaches the dew point.
        a = 1.5

        # Dampening constant for wind speed. This value ensures a gradual decay in fog density
        # as wind speed increases, which reflects that moderate wind disperses fog.
        b = 10

        # Calculate initial fog density based on temperature and humidity
        humidity_factor = rh2m / 100  # Relative humidity as a fraction
        temp_dew_diff = temp2m - dewpoint2m

        # Calculate fog density
        fog_density = k * humidity_factor * math.exp(-temp_dew_diff / a)

        # Adjust for wind speed
        adjusted_fog_density = fog_density * math.exp(-wind_speed / b)

        # Ensure fog density is within the valid range [0, 1]
        adjusted_fog_density = max(0, min(1, adjusted_fog_density))

        return float(adjusted_fog_density)

    # @typechecked
    # async def calculate_fog_density(self, temp2m, rh2m, dewpoint2m, wind_speed) -> int:
    #     ffp = max(
    #         0, 1 - abs(temp2m - dewpoint2m) / 5
    #     )  # Scale factor: closer to 1 if near saturation

    #     # (Relative Humidity Factor): Based on relative humidity; 1 at 100% RH, lower otherwise.
    #     rh_factor = min(1, rh2m / 100)  # 1 if RH is 100%, scales down otherwise

    #     # (Wind Influence): Decreases with increasing wind_speed, favoring calm conditions.
    #     wind_factor = max(0, 1 - wind_speed / 5)  # Scales down with higher wind speeds

    #     # Calculate Fog Density
    #     fog2m = ffp * rh_factor * wind_factor

    #     return fog2m

    # #####################################################
    # Calculate dew point at 2m
    # #####################################################
    @typechecked
    async def calculate_dew_point(self, temp2m, rh2m) -> float:
        """
        Calculate the dew point temperature given air temperature and relative humidity
        using Magnus formula.

        Parameters:
            temp (float): Air temperature in °C.
            humidity (float): Relative humidity in % (0-100).

        Args:
        - temp2m: Surface temperature in Celsius.
        - rh2m: Humidity in Percent.

        Returns:
        - float: Dew point temperature in °C.
        """
        # Constants
        a = 17.62
        b = 243.12

        # Calculate alpha
        alpha = math.log(rh2m / 100) + (a * temp2m) / (b + temp2m)

        return (b * alpha) / (a - alpha)

    # #####################################################
    # Atmospheric calculations
    # #####################################################
    @typechecked
    def _calculate_adjusted_pressure(self, pressure_sea_level, altitude) -> float:
        """
        Calculate the adjusted pressure at a given altitude above sea level.
        """

        lapse_rate = -0.0065  # Temperature lapse rate in K/m
        temperature_sea_level = 288.15  # Temperature at sea level in K
        gravity = 9.80665  # Acceleration due to gravity in m/s^2
        molar_mass_air = 0.02896  # Molar mass of Earth's air in kg/mol
        gas_constant = 8.31447  # Universal gas constant in J/(mol*K)

        pressure_adjusted = pressure_sea_level * (1 - (lapse_rate * altitude) / temperature_sea_level) ** (
            (gravity * molar_mass_air) / (gas_constant * lapse_rate)
        )

        return pressure_adjusted

    @typechecked
    def _calculate_vapor_pressure(self, temperature) -> float:
        """
        Calculate the actual or saturation vapor pressure at the surface using the Magnus-Tetens formula.
        If the surface temperature is given, the actual vapor pressure is calculated.
        The dew point temperature calculates ths saturation vapor pressure.

        Used by: lifted index, seeing

        Args:
        - temperature: Surface or dew point temperature in Celsius.

        Returns:
        - e: Actual vapor pressure at the surface in millibars (mb) or hectopascals (hPa).

        Magnus coefficients:

        Temperatures in between -45°C bis 60°C:
            a=17.62
            b=243.12
        Temperatures in between 0°C bis 100°C:
            a=7.5
            b=237.7
        Temperatures in between -10°C bis 50°C:
            a=17.27
            b=237.7
        """
        # # Constants
        # A = 17.62
        # B = 243.12

        # # Monteith and Unsworth (2008) provide Tetens' formula for temperatures above 0 °C
        A = 17.27
        B = 237.3

        # Murray (1967) provides Tetens' equation for temperatures below 0 °C
        # A = 21.875
        # B = 265.5

        # Calculate vapor pressure using Magnus-Tetens formula
        e = 0.61078 * math.exp((A * temperature) / (temperature + B))

        return e

    @typechecked
    def _calculate_water_vapor_pressure(self, dew_point_temperature, humidity) -> float:
        """
        Calculate the water vapor pressure based on temperature and humidity.

        Used by: seeing

        Args:
        - dew_point_temperature: Dew point temperature in Celsius.
        - humidity: Humidity in Percent.

        Returns:
        - water_vapor_pressure: Water vapor pressure at the surface in millibars (mb) or hectopascals (hPa).
        """

        es = self._calculate_vapor_pressure(dew_point_temperature)
        water_vapor_pressure = (humidity / 100) * es

        return water_vapor_pressure

    @typechecked
    def _calculate_mixing_ratio(self, e, air_pressure_at_sea_level) -> float:
        """
        Calculate the mixing ratio.

        Used by: lifted index

        Args:
        - e: Vapor pressure
        - air_pressure_at_sea_level: Air pressure at sea level

        Returns:
        - w: mixing ratio at surface in grams per kilogram
        """

        # Constants
        A = 621.97

        # Calculate actual vapor pressure using Magnus-Tetens formula
        w = A * (e / (air_pressure_at_sea_level - e))

        return w

    @typechecked
    def _calculate_lifting_condensation_level(self, w, air_pressure_at_sea_level) -> float:
        """
        The Lifting Condensation Level is the level at which a parcel becomes saturated. It can be used as
        a reasonable estimate of cloud base height when parcels experience forced ascent.
        Here, we calculate it using the Clausius-Clapeyron equation. This equation relates the saturation vapor
        pressure (e) to the temperature and pressure.
        https://en.wikipedia.org/wiki/Lifted_condensation_level

        Used by: lifted index

        Args:
        - w: Mixing ratio
        - air_pressure_at_sea_level: Air pressure at sea level

        Returns:
        - lcl: LCL in meters
        """

        # Constants
        A = 2440
        B = 0.00029

        # Calculate Lifting Condensation Level using the Clausius-Clapeyron equation
        lcl = (A * w) / ((air_pressure_at_sea_level - w) * (1 - B * air_pressure_at_sea_level))

        return lcl

    @typechecked
    def _calculate_transparency(
        self,
        humidity,
        temperature,
        cloud_cover,
        wind_speed,
        altitude,
        dew_point_temperature,
        air_pressure_at_sea_level,
        lifted_index,
        seeing,
    ) -> float:
        """
        The transparency of the atmosphere calculated by a weighted sum of conditions.

        Used by: magnitude degradation

        Args:
        - humidity: Humidity in Percent.
        - temperature: Surface temperature in Celsius.
        - cloud_cover: Cloud cover in Percent.
        - wind_speed: Wind speed in meters per second.
        - altitude: Altitude in meters.
        - dew_point_temperature: Dew point temperature in Celsius.
        - air_pressure_at_sea_level: Air pressure at sea level.
        - lifted_index: Lifted index in Celsius.
        - seeing: Seeing in Arcsecs.
        - w: Mixing ratio

        Returns:
        - transparency: In the range of 0 to 1
        """

        # Coefficients for the linear model (you can adjust these based on your requirements)
        temp_weight = 0.10
        humidity_weight = 0.10
        wind_weight = 0.10
        dew_point_weight = 0.10
        lifted_index_weight = 0.15
        seeing_weight = 0.15
        pressure_weight = 0.10
        cloud_cover_weight = 0.10
        altitude_weight = 0.10

        # Normalize values (optional)
        # Assuming typical temperature range from -10°C to 40°C
        temp_normalized = (temperature - (-10)) / (40 - (-10))
        humidity_normalized = humidity / 100.0
        # Assuming maximum typical wind speed as 20 m/s
        wind_normalized = wind_speed / 20.0
        # Assuming typical dew point range from -10°C to 30°C
        dew_point_normalized = (dew_point_temperature - (-10)) / (30 - (-10))
        # Assuming lifted index range from -8 to 8
        lifted_index_normalized = (lifted_index + 8) / 16.0
        # Assuming typical seeing conditions from 0 to 10 arcseconds
        seeing_normalized = (10 - seeing) / 10.0
        # Assuming typical pressure range from 900 to 1100 mb
        pressure_normalized = (air_pressure_at_sea_level - 900) / (1100 - 900)
        cloud_cover_normalized = cloud_cover / 100.0
        # Assuming typical altitude range from 0 to 5000 meters
        altitude_normalized = altitude / 5000.0

        transparency = 1 - (
            temp_weight * temp_normalized
            + humidity_weight * humidity_normalized
            + wind_weight * wind_normalized
            + dew_point_weight * dew_point_normalized
            + lifted_index_weight * lifted_index_normalized
            + seeing_weight * seeing_normalized
            + pressure_weight * pressure_normalized
            + cloud_cover_weight * cloud_cover_normalized
            + altitude_weight * altitude_normalized
        )

        # Ensure transparency is within the valid range [0, 1]
        transparency = max(0, min(1, transparency))

        return transparency

    @typechecked
    def _transparency_to_magnitude_degradation(self, transparency) -> float:
        """
        The magnitude degradation based on the atmospheric transparency.

        Used by: magnitude degradation

        Args:
        - transparency: In the range of 0 to 1.

        Returns:
        - magniture_degradation: In magnitude.
        """

        if transparency == 0:
            return float(MAG_DEGRATION_MAX)
        magnitude_degradation = -MAG_DEGRATION_MAX * math.log(transparency)
        magnitude_degradation = max(0, min(MAG_DEGRATION_MAX, magnitude_degradation))

        return magnitude_degradation


class AstronomicalRoutines:
    """Calculate different astronomical objects."""

    def __init__(
        self,
        location_data,
        forecast_time=None,
    ) -> None:
        self._location_data = location_data
        self._test_mode = False

        _LOGGER.debug("Timezone: %s", self._location_data.timezone_info)
        if forecast_time is None:
            self._forecast_time = datetime.now(UTC).replace(tzinfo=UTC)
        else:
            self._forecast_time = forecast_time.replace(tzinfo=UTC)
            self._test_mode = True

        self._sun_observer = None
        self._sun_observer_nautical = None
        self._sun_observer_astro = None
        self._moon_observer = None
        self._sun = None
        self._moon = None
        self._sun_data = {}
        self._moon_data = {}
        self._darkness_data = {}

        # Internal only
        self._sun_previous_rising_astro = None
        self._sun_previous_setting_astro = None

    def _test_data(self, data, keys) -> bool:
        """Test that specific values in a dictionary are not None"""

        if not all(data[key] is not None for key in keys):
            return False
        return True

    def utc_to_local_diff(self) -> float:
        """Returns the UTC Offset."""

        # Get the current time in the specified timezone
        now = datetime.now(ZoneInfo(self._location_data.timezone_info))

        # Get the offset in seconds
        offset_seconds = now.utcoffset().total_seconds()

        # Convert the offset to hours
        return offset_seconds / 3600

    async def time_shift(self) -> float:
        """Returns the time_shift to UTC in hours."""

        return int(self.utc_to_local_diff() * 3600)

    async def need_update(self, forecast_time=None) -> None:
        """Update Sun and Moon."""

        if forecast_time is not None:
            self._forecast_time = forecast_time.replace(tzinfo=UTC)

        _LOGGER.debug("Astronomical calculations updating")
        self._calculate_sun()
        self._calculate_moon()

    #
    # Observers
    #
    @typechecked
    def _get_sun_observer(self, below_horizon=ASTRONOMICAL_DUSK_DAWN) -> ephem.Observer:
        """Retrieves the ephem sun observer for the current location."""

        observer = ephem.Observer()
        observer.lon = str(self._location_data.longitude)  # * degree
        observer.lat = str(self._location_data.latitude)  # * degree
        observer.elevation = self._location_data.elevation
        observer.horizon = below_horizon * degree
        observer.pressure = 0
        observer.epoch = datetime.now().strftime("%Y/%m/%d")

        return observer

    @typechecked
    def _get_moon_observer(self) -> ephem.Observer:
        """Retrieves the ephem mon observer for the current location."""

        observer = ephem.Observer()
        observer.lon = str(self._location_data.longitude)  # * degree
        observer.lat = str(self._location_data.latitude)  # * degree
        observer.elevation = self._location_data.elevation
        # Naval Observatory Risings and Settings
        # Set horizon to minus 34 arcminutes
        # https://aa.usno.navy.mil/data/RS_OneDay
        observer.horizon = "-0:34"
        observer.pressure = 0
        observer.epoch = datetime.now().strftime("%Y/%m/%d")

        return observer

    # #########################################################################
    # Sun
    # #########################################################################
    @typechecked
    async def sun_data(self) -> SunData:
        """Returns sun data."""

        sd = SunDataModel(self._sun_data)
        try:
            return SunData(data=sd)
        except TypeError as ve:
            _LOGGER.error(f"Failed to parse Sun data: {self._sun_data}")
            _LOGGER.error(ve)
            return None

    async def sun_next_rising(self) -> datetime:
        """Returns sun next rising."""

        if (
            not self._test_data(
                self._sun_data,
                ["next_rising_astro", "next_rising_nautical", "next_rising_civil"],
            )
            or self._forecast_time > self._sun_data["next_rising_astro"]
            or self._forecast_time > self._sun_data["next_rising_nautical"]
            or self._forecast_time > self._sun_data["next_rising_civil"]
        ):
            _LOGGER.debug("Astronomical calculations updating sun_next_rising")
            self._calculate_sun()

        if self._sun_data.get("next_rising_astro", None) is not None:
            return self._sun_data["next_rising_astro"]
        if self._sun_data.get("next_rising_nautical", None) is not None:
            return self._sun_data["next_rising_nautical"]
        if self._sun_data.get("next_rising_civil", None) is not None:
            return self._sun_data["next_rising_civil"]

    async def sun_next_setting(self) -> datetime:
        """Returns sun next setting."""

        if (
            not self._test_data(
                self._sun_data,
                ["next_rising_astro", "next_rising_nautical", "next_rising_civil"],
            )
            or self._forecast_time > self._sun_data["next_setting_astro"]
            or self._forecast_time > self._sun_data["next_setting_nautical"]
            or self._forecast_time > self._sun_data["next_setting_civil"]
        ):
            _LOGGER.debug("Astronomical calculations updating sun_next_setting")
            self._calculate_sun()

        if self._sun_data.get("next_setting_astro", None) is not None:
            return self._sun_data["next_setting_astro"]
        if self._sun_data.get("next_setting_nautical", None) is not None:
            return self._sun_data["next_setting_nautical"]
        if self._sun_data.get("next_setting_civil", None) is not None:
            return self._sun_data["next_setting_civil"]

    def _calculate_sun(self) -> None:
        """Calculates sun risings and settings."""

        if self._sun_observer is None:
            self._sun_observer = self._get_sun_observer(CIVIL_DUSK_DAWN)
        if self._sun_observer_nautical is None:
            self._sun_observer_nautical = self._get_sun_observer(NAUTICAL_DUSK_DAWN)
        if self._sun_observer_astro is None:
            self._sun_observer_astro = self._get_sun_observer(ASTRONOMICAL_DUSK_DAWN)
        if self._sun is None:
            self._sun = ephem.Sun()

        self._calculate_sun_civil()
        self._calculate_sun_nautical()
        self._calculate_sun_astro()
        self._calculate_sun_altaz()
        self._calculate_sun_constellation()

    def _calculate_sun_civil(self) -> None:
        # Rise and Setting (Civil)
        try:
            self._sun_data["next_rising_civil"] = (
                self._sun_observer.next_rising(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next rising
            start = self._sun_observer.date.datetime()
            end = self._sun_observer.date.datetime() + timedelta(days=365)
            timestamp = start
            cnt = 1
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer.date = timestamp
                try:
                    self._sun_data["next_rising_civil"] = (
                        self._sun_observer.next_rising(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    cnt += 1
                    continue
                _LOGGER.debug(f"Sun next rising civil in {cnt} days.")
                break

        try:
            self._sun_data["next_setting_civil"] = (
                self._sun_observer.next_setting(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next setting
            start = self._sun_observer.date.datetime()
            end = self._sun_observer.date.datetime() + timedelta(days=365)
            timestamp = start
            cnt = 1
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer.date = timestamp
                try:
                    self._sun_data["next_setting_civil"] = (
                        self._sun_observer.next_setting(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    cnt += 1
                    continue
                _LOGGER.debug(f"Sun next setting civil in {cnt} days.")
                break

    def _calculate_sun_nautical(self) -> None:
        # Rise and Setting (Nautical)
        self._sun_observer_nautical.date = self._forecast_time
        self._sun.compute(self._sun_observer_nautical)

        try:
            self._sun_data["next_rising_nautical"] = (
                self._sun_observer_nautical.next_rising(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next astronomical rising
            start = self._sun_observer_nautical.date.datetime()
            end = self._sun_observer_nautical.date.datetime() + timedelta(days=365)
            timestamp = start
            cnt = 1
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer_nautical.date = timestamp
                try:
                    self._sun_data["next_rising_nautical"] = (
                        self._sun_observer_nautical.next_rising(ephem.Sun(), use_center=True)
                        .datetime()
                        .replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    cnt += 1
                    continue
                _LOGGER.debug(f"Sun next rising nautical in {cnt} days.")
                break

        try:
            self._sun_data["next_setting_nautical"] = (
                self._sun_observer_nautical.next_setting(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next astronomical setting
            start = self._sun_observer_nautical.date.datetime()
            end = self._sun_observer_nautical.date.datetime() + timedelta(days=365)
            timestamp = start
            cnt = 1
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer_nautical.date = timestamp
                try:
                    self._sun_data["next_setting_nautical"] = (
                        self._sun_observer_nautical.next_setting(ephem.Sun(), use_center=True)
                        .datetime()
                        .replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    cnt += 1
                    continue
                _LOGGER.debug(f"Sun next setting nautical in {cnt} days.")
                break

    def _calculate_sun_astro(self) -> None:
        # Rise and Setting (Astronomical)
        self._sun_observer_astro.date = self._forecast_time
        self._sun.compute(self._sun_observer_astro)

        try:
            self._sun_data["next_rising_astro"] = (
                self._sun_observer_astro.next_rising(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next astronomical rising
            start = self._sun_observer_astro.date.datetime()
            end = self._sun_observer_astro.date.datetime() + timedelta(days=365)
            timestamp = start
            cnt = 1
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer_astro.date = timestamp
                try:
                    self._sun_data["next_rising_astro"] = (
                        self._sun_observer_astro.next_rising(ephem.Sun(), use_center=True)
                        .datetime()
                        .replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    cnt += 1
                    continue
                _LOGGER.debug(f"Sun next rising astro in {cnt} days.")
                break

        try:
            self._sun_data["previous_rising_astro"] = (
                self._sun_observer_astro.previous_rising(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the previous astronomical rising
            start = self._sun_observer_astro.date.datetime()
            end = self._sun_observer_astro.date.datetime() - timedelta(days=365)
            timestamp = start
            cnt = 1
            while timestamp > end:
                timestamp -= timedelta(minutes=1440)
                self._sun_observer_astro.date = timestamp
                try:
                    self._sun_data["previous_rising_astro"] = (
                        self._sun_observer_astro.previous_rising(ephem.Sun(), use_center=True)
                        .datetime()
                        .replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    cnt += 1
                    continue
                _LOGGER.debug(f"Sun previous rising astro {cnt} days before.")
                break

        try:
            self._sun_data["next_setting_astro"] = (
                self._sun_observer_astro.next_setting(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next astronomical setting
            start = self._sun_observer_astro.date.datetime()
            end = self._sun_observer_astro.date.datetime() + timedelta(days=365)
            timestamp = start
            cnt = 1
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer_astro.date = timestamp
                try:
                    self._sun_data["next_setting_astro"] = (
                        self._sun_observer_astro.next_setting(ephem.Sun(), use_center=True)
                        .datetime()
                        .replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    cnt += 1
                    continue
                _LOGGER.debug(f"Sun next setting astro in {cnt} days.")
                break

        try:
            self._sun_data["previous_setting_astro"] = (
                self._sun_observer_astro.previous_setting(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the previous astronomical setting
            start = self._sun_observer_astro.date.datetime()
            end = self._sun_observer_astro.date.datetime() - timedelta(days=365)
            timestamp = start
            cnt = 1
            while timestamp > end:
                timestamp -= timedelta(minutes=1440)
                self._sun_observer_astro.date = timestamp
                try:
                    self._sun_data["previous_setting_astro"] = (
                        self._sun_observer_astro.previous_setting(ephem.Sun(), use_center=True)
                        .datetime()
                        .replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    cnt += 1
                    continue
                _LOGGER.debug(f"Sun previous setting astro {cnt} days before.")
                break

    def _calculate_sun_altaz(self) -> None:
        """Calculates sun altitude and azimuth."""

        if self._sun_observer is None:
            self._sun_observer = self._get_sun_observer(CIVIL_DUSK_DAWN)
        if self._sun is None:
            self._sun = ephem.Sun()

        self._sun_observer.date = self._forecast_time
        self._sun.compute(self._sun_observer)

        # Sun Altitude
        self._sun_data["altitude"] = deg(float(self._sun.alt))

        # Sun Azimuth
        self._sun_data["azimuth"] = deg(float(self._sun.az))

    def _calculate_sun_constellation(self) -> None:
        """Calculates sun altitude and azimuth."""

        if self._sun_observer is None:
            self._sun_observer = self._get_sun_observer(CIVIL_DUSK_DAWN)
        if self._sun is None:
            self._sun = ephem.Sun()

        self._sun_observer.date = self._forecast_time
        self._sun.compute(self._sun_observer)

        # Sun Constellation
        constellation = ephem.constellation(self._sun)[1]
        self._sun_data["constellation"] = constellation

    # #########################################################################
    # Moon
    # #########################################################################
    @typechecked
    async def moon_data(self) -> MoonData:
        """Returns moon data."""

        md = MoonDataModel(self._moon_data)
        try:
            return MoonData(data=md)
        except TypeError as ve:
            _LOGGER.error(f"Failed to parse Moon data: {self._moon_data}")
            _LOGGER.error(ve)
            return None

    def _calculate_moon(self) -> None:
        """Calculates moon rising and setting."""

        if self._moon_observer is None:
            self._moon_observer = self._get_moon_observer()
        if self._moon is None:
            self._moon = ephem.Moon()

        # Rise and Setting
        self._moon_observer.date = self._forecast_time
        self._moon.compute(self._moon_observer)

        try:
            self._moon_data["next_rising"] = (
                self._moon_observer.next_rising(ephem.Moon()).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next astronomical rising
            start = self._moon_observer.date.datetime()
            end = self._moon_observer.date.datetime() + timedelta(days=365)
            timestamp = start
            cnt = 1
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._moon_observer.date = timestamp
                try:
                    self._moon_data["next_rising"] = (
                        self._moon_observer.next_rising(ephem.Moon()).datetime().replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    cnt += 1
                    continue
                _LOGGER.debug(f"Moon next rising in {cnt} days.")
                break

        try:
            self._moon_data["next_setting"] = (
                self._moon_observer.next_setting(ephem.Moon()).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next astronomical setting
            start = self._moon_observer.date.datetime()
            end = self._moon_observer.date.datetime() + timedelta(days=365)
            timestamp = start
            cnt = 1
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._moon_observer.date = timestamp
                try:
                    self._moon_data["next_setting"] = (
                        self._moon_observer.next_setting(ephem.Moon()).datetime().replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    cnt += 1
                    continue
                _LOGGER.debug(f"Moon next setting in {cnt} days.")
                break

        try:
            self._moon_data["previous_rising"] = (
                self._moon_observer.previous_rising(ephem.Moon()).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the previous astronomical rising
            start = self._moon_observer.date.datetime()
            end = self._moon_observer.date.datetime() - timedelta(days=365)
            timestamp = start
            cnt = 1
            while timestamp > end:
                timestamp -= timedelta(minutes=1440)
                self._moon_observer.date = timestamp
                try:
                    self._moon_data["previous_rising"] = (
                        self._moon_observer.previous_rising(ephem.Moon()).datetime().replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    cnt += 1
                    continue
                _LOGGER.debug(f"Moon previous rising {cnt} days before.")
                break

        try:
            self._moon_data["previous_setting"] = (
                self._moon_observer.previous_setting(ephem.Moon()).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the previous astronomical setting
            start = self._moon_observer.date.datetime()
            end = self._moon_observer.date.datetime() - timedelta(days=365)
            timestamp = start
            cnt = 1
            while timestamp > end:
                timestamp -= timedelta(minutes=1440)
                self._moon_observer.date = timestamp
                try:
                    self._moon_data["previous_setting"] = (
                        self._moon_observer.previous_setting(ephem.Moon()).datetime().replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    cnt += 1
                    continue
                _LOGGER.debug(f"Moon previous setting {cnt} days before.")
                break

        # self._moon_observer.date = self._forecast_time + timedelta(days=1)
        # self._moon.compute(self._moon_observer)

        # try:
        #     self._moon_day_after_next_rising = self._moon_observer.next_rising(ephem.Moon()).datetime().replace(tzinfo=UTC)
        # except (ephem.AlwaysUpError, ephem.NeverUpError):
        #     pass

        # try:
        #     self._moon_day_after_next_setting = self._moon_observer.next_setting(ephem.Moon()).datetime().replace(tzinfo=UTC)
        # except (ephem.AlwaysUpError, ephem.NeverUpError):
        #     pass

        # Next new Moon
        self._moon_data["next_new_moon"] = ephem.next_new_moon(self._forecast_time).datetime().replace(tzinfo=UTC)

        # Pervious new Moon
        self._moon_data["previous_new_moon"] = (
            ephem.previous_new_moon(self._forecast_time).datetime().replace(tzinfo=UTC)
        )

        # Next full Moon
        self._moon_data["next_full_moon"] = ephem.next_full_moon(self._forecast_time).datetime().replace(tzinfo=UTC)

        # Moon phase
        self._moon_data["phase"] = self._moon.phase

        self._calculate_moon_altaz()
        self._calculate_moon_distance_size()
        self._calculate_moon_constellation()
        self._calculate_moon_phase_on_day()
        self._next_dark_night()

    def _calculate_moon_altaz(self) -> None:
        """Calculates moon altitude and azimuth."""

        if self._moon_observer is None:
            self._moon_observer = self._get_moon_observer()
        if self._moon is None:
            self._moon = ephem.Moon()

        self._moon_observer.date = self._forecast_time
        self._moon.compute(self._moon_observer)

        # Moon Altitude
        self._moon_data["altitude"] = deg(float(self._moon.alt))

        # Moon Azimuth
        self._moon_data["azimuth"] = deg(float(self._moon.az))

    def _calculate_moon_distance_size(self) -> None:
        """Calculate moon distance and relative size"""

        if self._moon_observer is None:
            self._moon_observer = self._get_moon_observer()
        if self._moon is None:
            self._moon = ephem.Moon()

        # Get the distance in Earth radii
        self._moon_data["distance"] = self._moon.earth_distance  # in AU (Astronomical Units)

        # Convert to kilometers
        self._moon_data["distance_km"] = self._moon_data["distance"] * 149597870.7  # 1 AU = 149597870.7 km

        # Get the Moon's angular size (in degrees)
        self._moon_data["angular_size"] = self._moon.radius * 2 * 57.29578  # 180 / pi

        # Average distance and angular size for comparison
        self._moon_data["avg_distance_km"] = 384400  # Average distance of the Moon from Earth in km
        self._moon_data["avg_angular_size"] = 0.5181  # Average angular size in degrees

        # Relative distance and size compared to average
        self._moon_data["relative_distance"] = self._moon_data["distance_km"] / self._moon_data["avg_distance_km"]
        self._moon_data["relative_size"] = self._moon_data["angular_size"] / self._moon_data["avg_angular_size"]

    def _calculate_moon_constellation(self) -> None:
        """Calculates moon constellation."""

        if self._moon_observer is None:
            self._moon_observer = self._get_moon_observer()
        if self._moon is None:
            self._moon = ephem.Moon()

        self._moon_observer.date = self._forecast_time
        self._moon.compute(self._moon_observer)

        # Moon Constellation
        constellation = ephem.constellation(self._moon)[1]
        self._moon_data["constellation"] = constellation

    def _calculate_moon_phase_on_day(self) -> None:
        """
        Calculates the Moon Phase on day

        The following extract the percent time between one new moon and the next
        This corresponds (somewhat roughly) to the phase of the moon.
        Note that there is a ephem.Moon().phase(), but this returns the
        percentage of the moon which is illuminated. This is not really what we
        want.
        """

        day = 1.0 / 29.33
        moonphase = [
            (0.0 / 4.0 + day, "🌑", "New moon", "moon-new"),
            (1.0 / 4.0 - day, "🌒", "Waxing crescent moon", "moon-waxing-crescent"),
            (1.0 / 4.0 + day, "🌓", "First quarter moon", "first-quarter-moon"),
            (2.0 / 4.0 - day, "🌔", "Waxing gibbous moon", "moon-waxing-gibbous"),
            (2.0 / 4.0 + day, "🌕", "Full moon", "moon-full"),
            (3.0 / 4.0 - day, "🌖", "Waning gibbous moon", "moon-waning-gibbous"),
            (3.0 / 4.0 + day, "🌗", "Last quarter moon", "moon-last-quarter"),
            (4.0 / 4.0 - day, "🌘", "Waning crescent moon", "moon-waning-crescent"),
            (4.0 / 4.0, "🌑", "New moon", "moon-new"),
        ]
        ranges = [x[0] for x in moonphase]

        nnm = self._moon_data["next_new_moon"]
        pnm = self._moon_data["previous_new_moon"]

        # Number from 0-1. where 0=new, 0.5=full, 1=new
        lunation = (self._forecast_time - pnm) / (nnm - pnm)

        phase = bisect.bisect_right(ranges, lunation)

        self._moon_data["icon"] = moonphase[phase][3]

        # _LOGGER.debug(f"{round(lunation * 100.0, 1)} {moonphase[phase][1]} {moonphase[phase][2]} {moonphase[phase][3]}")

    def _get_astronomical_darkness(self, date) -> tuple | tuple[None, None]:
        """
        Get astronimical darkness for a given date.

        Args:
        - date: Date for calculation

        Returns:
        - astro_twilight_start,
          astro_twilight_end: Datetime for astronomical sun set and rise.
                              If the sun does not rise or set [None, None] is returned.
        """
        self._sun_observer_astro.date = date

        # Compute the next sunset and sunrise
        try:
            astro_twilight_start = self._sun_observer_astro.next_setting(ephem.Sun(), use_center=True)
            self._sun_observer_astro.date = astro_twilight_start
            astro_twilight_end = self._sun_observer_astro.next_rising(ephem.Sun(), use_center=True)
            return astro_twilight_start.datetime(), astro_twilight_end.datetime()
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            return None, None

    # Function to check Moon conditions for the whole night
    def _is_moon_dark_whole_night(self, night_start, night_end) -> bool:
        """
        Test if Moon phase is less than 5% and/or maximum Moon altitide during
        astronomical darkness is less than 5°.

        Args:
        - night_start: Sun set -18°
        - night_end: Sun rise -18°

        Returns:
        - dark_night: True, if Moon is within constraints.
        """
        current_time = night_start

        moon_alts = []
        moon_phases = []
        dark_night = True

        while current_time <= night_end:
            self._moon_observer.date = current_time
            moon = ephem.Moon(self._moon_observer)

            moon_alt = moon.alt
            moon_alts.append(moon_alt)
            moon_phase = moon.phase  # Moon phase in percentage
            moon_phases.append(moon_phase)

            # If at any point the Moon is above horizon and > 10% illuminated, return False
            if moon_alt > ephem.degrees(DARK_NIGHT_MAX_MOON_ALT):  # and moon_phase >= DARK_NIGHT_MAX_MOON_PHASE * 10:
                dark_night = False
                break

            # Step in time (e.g., every 15 minutes)
            current_time += timedelta(minutes=5)

        moon_phases_avg = sum(moon_phases) / len(moon_phases)
        if dark_night or moon_phases_avg <= DARK_NIGHT_MAX_MOON_PHASE * 10:
            _LOGGER.debug(
                f"Next dark night: {night_start.strftime("%Y-%m-%d")}, Altidudes min/max: {min(moon_alts)}/{max(moon_alts)}, Phase mix/max: {min(moon_phases):.2f}/{max(moon_phases):.2f}%"
            )
        else:
            dark_night = False
        return dark_night

    def _next_dark_night(self) -> None:
        """Calculate the next dark night"""

        if self._moon_observer is None:
            self._moon_observer = self._get_moon_observer()
        if self._moon is None:
            self._moon = ephem.Moon()

        self._moon_data["next_dark_night"] = None

        start_date = datetime.now(UTC).replace(hour=12, minute=0, second=0, microsecond=0) - timedelta(
            seconds=self.utc_to_local_diff()
        )

        for i in range(365):
            date = start_date + timedelta(days=i)
            night_start, night_end = self._get_astronomical_darkness(date)

            # Skip nights where twilight is undefined (polar regions)
            if night_start is None or night_end is None:
                continue

            if self._is_moon_dark_whole_night(night_start, night_end):
                self._moon_data["next_dark_night"] = night_start.replace(tzinfo=UTC)
                break

    # #########################################################################
    # Darkness
    # #########################################################################
    @typechecked
    async def darkness_data(self) -> DarknessData:
        """Returns darkness data."""

        self._darkness_data["deep_sky_darkness_moon_rises"] = self._deep_sky_darkness_moon_rises()
        self._darkness_data["deep_sky_darkness_moon_sets"] = self._deep_sky_darkness_moon_sets()
        self._darkness_data["deep_sky_darkness_moon_always_up"] = self._deep_sky_darkness_moon_always_up()
        self._darkness_data["deep_sky_darkness_moon_always_down"] = self._deep_sky_darkness_moon_always_down()
        self._darkness_data["deep_sky_darkness"] = self._deep_sky_darkness()

        dd = DarknessDataModel(self._darkness_data)
        try:
            return DarknessData(data=dd)
        except TypeError as ve:
            _LOGGER.error(f"Failed to parse darkness data: {self._darkness_data}")
            _LOGGER.error(ve)
            return None

    async def night_duration_astronomical(self) -> float:
        """Returns the remaining timespan of astronomical darkness."""

        start_timestamp = None

        # Are we already in darkness?
        if self._astronomical_darkness():
            start_timestamp = self._sun_data["previous_setting_astro"]
        else:
            start_timestamp = self._sun_data["next_setting_astro"]

        return (self._sun_data["next_rising_astro"] - start_timestamp).total_seconds()

    def _deep_sky_darkness(self) -> float:
        """Returns the remaining timespan of deep sky darkness."""

        dsd = timedelta(0)

        if self._astronomical_darkness():
            _LOGGER.debug("DSD: In astronomical darkness")
            if self._deep_sky_darkness_moon_rises():
                dsd = self._moon_data["next_rising"] - self._forecast_time
                _LOGGER.debug(f"DSD: Sun down, Moon rises {dsd}")

            if self._deep_sky_darkness_moon_sets():
                if self._moon_down():
                    dsd = self._sun_data["next_rising_astro"] - self._forecast_time
                    _LOGGER.debug(f"DSD: Sun down, Moon is down {dsd}")
                else:
                    dsd = self._sun_data["next_rising_astro"] - self._moon_data["next_setting"]
                    _LOGGER.debug(f"DSD: Sun down, Moon sets {dsd}")

            if self._deep_sky_darkness_moon_always_down():
                dsd = self._sun_data["next_rising_astro"] - self._forecast_time
                _LOGGER.debug(f"DSD: Moon always down {dsd}")
            else:
                _LOGGER.debug(f"DSD: Moon NOT always down {dsd}")

        if not self._astronomical_darkness():
            _LOGGER.debug("DSD: At sunlight")
            if self._deep_sky_darkness_moon_rises():
                dsd = self._moon_data["next_rising"] - self._sun_data["next_setting_astro"]
                _LOGGER.debug(f"DSD: Sun up, Moon rises {dsd}")

            if self._deep_sky_darkness_moon_sets():
                dsd = self._sun_data["next_rising_astro"] - self._moon_data["next_setting"]
                _LOGGER.debug(f"DSD: Sun up, Moon sets {dsd}")

            if self._deep_sky_darkness_moon_always_down():
                dsd = self._sun_data["next_rising_astro"] - self._sun_data["next_setting_astro"]
                _LOGGER.debug(f"DSD: Sun up, Moon down {dsd}")

        if self._deep_sky_darkness_moon_always_up():
            dsd = timedelta(0)
            _LOGGER.debug(f"DSD: Moon always up {dsd}")
        else:
            _LOGGER.debug(f"DSD: Moon NOT always up {dsd}")

        return dsd.total_seconds()

    def _astronomical_darkness(self) -> bool:
        """Returns true during astronomical night."""

        if self._sun_data["next_setting_astro"] > self._sun_data["next_rising_astro"]:
            return True
        return False

    def _moon_down(self) -> bool:
        """Returns true while moon is set-"""

        if self._moon_data["next_setting"] > self._moon_data["next_rising"]:
            return True
        return False

    def _deep_sky_darkness_moon_rises(self) -> bool:
        """Returns true if moon rises during astronomical night."""

        start_timestamp = None

        # Are we already in darkness?
        if self._astronomical_darkness():
            start_timestamp = self._sun_data["previous_setting_astro"]
        else:
            start_timestamp = self._sun_data["next_setting_astro"]

        if (
            self._moon_data["next_rising"] > start_timestamp
            and self._moon_data["next_rising"] < self._sun_data["next_rising_astro"]
        ):
            _LOGGER.debug("DSD: Moon rises during astronomical night")
            return True
        return False

    def _deep_sky_darkness_moon_sets(self) -> bool:
        """Returns true if moon sets during astronomical night."""

        start_timestamp = None

        # Are we already in darkness?
        if self._astronomical_darkness():
            start_timestamp = self._sun_data["previous_setting_astro"]
        else:
            start_timestamp = self._sun_data["next_setting_astro"]

        # Did Moon already set in darkness?
        if self._moon_down() and self._astronomical_darkness():
            start_timestamp_moon = self._moon_data["previous_setting"]
        else:
            start_timestamp_moon = self._moon_data["next_setting"]

        if start_timestamp_moon > start_timestamp and start_timestamp_moon < self._sun_data["next_rising_astro"]:
            _LOGGER.debug("DSD: Moon sets during astronomical night")
            return True
        return False

    def _deep_sky_darkness_moon_always_up(self) -> bool:
        """Returns true if moon is up during astronomical night."""

        start_timestamp = None

        # Are we already in darkness?
        if self._astronomical_darkness():
            start_timestamp = self._sun_data["previous_setting_astro"]
        else:
            start_timestamp = self._sun_data["next_setting_astro"]

        if (
            self._moon_data["next_rising"] < start_timestamp
            and self._moon_data["next_setting"] > self._sun_data["next_rising_astro"]
        ):
            _LOGGER.debug("DSD: Moon is up during astronomical night")
            return True
        return False

    def _deep_sky_darkness_moon_always_down(self) -> bool:
        """Returns true if moon is down during astronomical night."""

        start_timestamp = None

        # Are we already in darkness?
        if self._astronomical_darkness():
            start_timestamp = self._sun_data["previous_setting_astro"]
        else:
            start_timestamp = self._sun_data["next_setting_astro"]

        if (
            self._moon_data["previous_setting"] < start_timestamp
            and self._moon_data["next_rising"] > self._sun_data["next_rising_astro"]
        ):
            _LOGGER.debug("DSD: Moon is down during astronomical night")
            return True
        return False



from setuptools import setup
import os

lib_folder = os.path.dirname(os.path.realpath(__file__))
requirement_path = f"{lib_folder}/requirements.txt"
install_requires = []
if os.path.isfile(requirement_path):
    with open(requirement_path) as f:
        install_requires = f.read().splitlines()

setup(
    name="pyastroweatherio",
    packages=["pyastroweatherio"],
    version="0.73.1.dev8",
    license="MIT",
    description="Python Wrapper for OpenMeteo and Met.no REST API",
    long_description=" ".join(
        [
            "Lightweight Python 3 module to receive data via",
            "REST API from OpenMeteo and Met.no and consume",
            "UpTonight reports.",
        ],
    ),
    author="Markus Winkler",
    author_email="winkler.info@icloud.com",
    url="https://github.com/mawinkler/pyastroweatherio",
    keywords=["AstroWeather", "Met.no", "OpenMeteo", "Python"],
    install_requires=install_requires,
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
)



"""Define a client to interact with weather APIs."""

import asyncio
import json
import logging
import math
import os.path
import socket
from datetime import UTC, datetime, timedelta, timezone
from json.decoder import JSONDecodeError
from pprint import pprint as pp
from typing import Any, Dict, List, Optional

import aiofiles

# from aiohttp.client import ClientError, ClientResponseError, ClientSession
import numpy as np  # from retry_requests import retry

# import openmeteo_requests
# import requests_cache
import pandas as pd
from aiohttp import ClientSession, ClientTimeout
from aiohttp.client import ClientResponseError, ClientSession
from aiohttp.client_exceptions import ClientError
from typeguard import typechecked

from pyastroweatherio.const import (
    BASE_URL_MET,
    BASE_URL_OPENMETEO,
    DEFAULT_CACHE_TIMEOUT,
    DEFAULT_CONDITION_CALM_WEIGHT,
    DEFAULT_CONDITION_CLOUDCOVER_HIGH_WEAKENING,
    DEFAULT_CONDITION_CLOUDCOVER_LOW_WEAKENING,
    DEFAULT_CONDITION_CLOUDCOVER_MEDIUM_WEAKENING,
    DEFAULT_CONDITION_CLOUDCOVER_WEIGHT,
    DEFAULT_CONDITION_FOG_WEIGHT,
    DEFAULT_CONDITION_SEEING_WEIGHT,
    DEFAULT_CONDITION_TRANSPARENCY_WEIGHT,
    DEFAULT_ELEVATION,
    DEFAULT_LATITUDE,
    DEFAULT_LONGITUDE,
    DEFAULT_TIMEOUT,
    DEFAULT_TIMEZONE,
    FORECAST_TYPE_HOURLY,
    HEADERS,
    MAG_DEGRATION_MAX,
    SEEING,
    SEEING_MAX,
    TRANSPARENCY,
    WIND10M_MAX,
)
from pyastroweatherio.dataclasses import (
    ConditionData,
    ConditionDataModel,
    ForecastData,
    ForecastDataModel,
    GeoLocationData,
    GeoLocationDataModel,
    LocationData,
    LocationDataModel,
    NightlyConditionsData,
    NightlyConditionsDataModel,
    TimeData,
    TimeDataModel,
    UpTonightBodiesData,
    UpTonightBodiesDataModel,
    UpTonightCometsData,
    UpTonightCometsDataModel,
    UpTonightDSOData,
    UpTonightDSODataModel,
)
from pyastroweatherio.errors import OpenMeteoConnectionError, OpenMeteoError
from pyastroweatherio.helper_functions import (
    AstronomicalRoutines,
    AtmosphericRoutines,
    ConversionFunctions,
)

_LOGGER = logging.getLogger(__name__)
_NOT_AVAILABLE = -9999


class AstroWeather:
    """AstroWeather Communication Client."""

    def __init__(
        self,
        session: Optional[ClientSession] = None,
        latitude=DEFAULT_LATITUDE,
        longitude=DEFAULT_LONGITUDE,
        elevation=DEFAULT_ELEVATION,
        timezone_info=DEFAULT_TIMEZONE,
        cloudcover_weight=DEFAULT_CONDITION_CLOUDCOVER_WEIGHT,
        cloudcover_high_weakening=DEFAULT_CONDITION_CLOUDCOVER_HIGH_WEAKENING,
        cloudcover_medium_weakening=DEFAULT_CONDITION_CLOUDCOVER_MEDIUM_WEAKENING,
        cloudcover_low_weakening=DEFAULT_CONDITION_CLOUDCOVER_LOW_WEAKENING,
        fog_weight=DEFAULT_CONDITION_FOG_WEIGHT,
        seeing_weight=DEFAULT_CONDITION_SEEING_WEIGHT,
        transparency_weight=DEFAULT_CONDITION_TRANSPARENCY_WEIGHT,
        calm_weight=DEFAULT_CONDITION_CALM_WEIGHT,
        uptonight_path="/conf/www",
        test_datetime=None,
        experimental_features=False,
        forecast_model=None,
    ):
        self._session: ClientSession = session
        self._location_data = self._get_location(
            latitude,
            longitude,
            elevation,
            timezone_info,
        )
        self._weather_data_uptonight = {}
        self._weather_data_uptonight_bodies = {}
        self._weather_data_uptonight_comets = {}
        self._weather_data_timestamp = datetime.now() - timedelta(seconds=(DEFAULT_CACHE_TIMEOUT + 1))
        self._cloudcover_weight = cloudcover_weight
        self._cloudcover_high_weakening = cloudcover_high_weakening
        self._cloudcover_medium_weakening = cloudcover_medium_weakening
        self._cloudcover_low_weakening = cloudcover_low_weakening
        self._fog_weight = fog_weight
        self._seeing_weight = seeing_weight
        self._transparency_weight = transparency_weight
        self._calm_weight = calm_weight
        self._uptonight_path = uptonight_path
        self._test_datetime = test_datetime
        self._experimental_features = True
        self._forecast_model = forecast_model
        # Open-Meteo
        # DWD Germany: icon_seamless
        # Met.no: metno_seamless
        # NOAA U.S.: gfs_seamless
        # ECMWF: ecmwf_ifs025

        # Weather data
        self._weather_df = None

        # Forecast data
        self._forecast_data = None

        # Astro Routines
        self._astro_routines = AstronomicalRoutines(
            self._location_data,
            self._test_datetime,
        )

        self._atmosphere = AtmosphericRoutines()

        self._lock = asyncio.Lock()
        
        # Testing
        self._test_mode = False
        if self._test_datetime is not None:
            self._test_mode = True

    # #########################################################################
    # Public functions
    # #########################################################################
    @typechecked
    async def get_location_data(
        self,
    ) -> List[LocationData]:
        """Returns station Weather Forecast."""

        return await self._get_location_data()

    @typechecked
    async def get_hourly_forecast(self) -> List[ForecastData]:
        """Returns hourly Weather Forecast."""

        return await self._get_forecast_data(FORECAST_TYPE_HOURLY, 72)

    @typechecked
    async def get_deepsky_forecast(self) -> List[NightlyConditionsData]:
        """Returns Deep Sky Forecast."""

        return await self._get_deepsky_forecast()

    # #########################################################################
    # Private functions
    # #########################################################################
    async def _calculate_dew_point(self, df):
        tasks = [
            self._atmosphere.calculate_dew_point(
                temp2m=row["air_temperature"],
                rh2m=row["relative_humidity"],
            )
            for _, row in df.iterrows()
        ]
        results = await asyncio.gather(*tasks)
        return results

    async def _calculate_seeing(self, df):
        tasks = [
            self._atmosphere.calculate_seeing(
                temperature=row["air_temperature"],
                humidity=row["relative_humidity"],
                dew_point_temperature=row["dew_point_temperature"],
                wind_speed=row["wind_speed"],
                cloud_cover=row["cloud_area_fraction"],
                altitude=self._location_data.elevation,
                air_pressure_at_sea_level=row["air_pressure_at_sea_level"],
            )
            for _, row in df.iterrows()
        ]
        results = await asyncio.gather(*tasks)
        return results

    async def _calculate_transparency(self, df):
        tasks = [
            self._atmosphere.magnitude_degradation(
                temperature=row["air_temperature"],
                humidity=row["relative_humidity"],
                dew_point_temperature=row["dew_point_temperature"],
                wind_speed=row["wind_speed"],
                cloud_cover=row["cloud_area_fraction"],
                altitude=self._location_data.elevation,
                air_pressure_at_sea_level=row["air_pressure_at_sea_level"],
            )
            for _, row in df.iterrows()
        ]
        results = await asyncio.gather(*tasks)
        return results

    async def _calculate_lifted_index(self, df):
        tasks = [
            self._atmosphere.calculate_lifted_index(
                temperature=row["air_temperature"],
                dew_point_temperature=row["dew_point_temperature"],
                altitude=self._location_data.elevation,
                air_pressure_at_sea_level=row["air_pressure_at_sea_level"],
            )
            for _, row in df.iterrows()
        ]
        results = await asyncio.gather(*tasks)
        return results

    async def _retrive_data(self) -> None:
        """Retrieves current data from all data sources."""
        async with self._lock:
            if ((datetime.now() - self._weather_data_timestamp).total_seconds()) > DEFAULT_CACHE_TIMEOUT:
                self._weather_data_timestamp = datetime.now()

                weather_df_metno = await self._retrieve_data_metno()
                if weather_df_metno is None:
                    _LOGGER.warning("Could not retrieve Met.no data. Using existing data.")
                    return None
                await self._retrieve_data_uptonight()

                weather_df_openmeteo = await self._retrieve_data_openmeteo()

                # Merge the dataframes of metno and openmeteo
                self._weather_df = weather_df_metno.merge(
                    weather_df_openmeteo, on="time", how="left"
                )

                # Overwrite met.no forecast with open-meteo forcast
                self._weather_df["cloud_area_fraction"] = self._weather_df["openmeteo_cloud_cover"]
                self._weather_df["cloud_area_fraction_high"] = self._weather_df["openmeteo_cloud_cover_high"]
                self._weather_df["cloud_area_fraction_medium"] = self._weather_df["openmeteo_cloud_cover_mid"]
                self._weather_df["cloud_area_fraction_low"] = self._weather_df["openmeteo_cloud_cover_low"]
                self._weather_df["relative_humidity"] = self._weather_df["openmeteo_relative_humidity_2m"]
                self._weather_df["wind_speed"] = self._weather_df["openmeteo_wind_speed_10m"]
                self._weather_df["wind_from_direction"] = self._weather_df["openmeteo_wind_direction_10m"]
                self._weather_df["air_temperature"] = self._weather_df["openmeteo_temperature_2m"]
                self._weather_df["dew_point_temperature"] = self._weather_df["openmeteo_dew_point_2m"]
                self._weather_df["precipitation_amount"] = self._weather_df["openmeteo_precipitation"]
                self._weather_df = self._weather_df.drop(
                    [
                        "openmeteo_cloud_cover",
                        "openmeteo_cloud_cover_high",
                        "openmeteo_cloud_cover_mid",
                        "openmeteo_cloud_cover_low",
                        "openmeteo_relative_humidity_2m",
                        "openmeteo_wind_speed_10m",
                        "openmeteo_wind_direction_10m",
                        "openmeteo_temperature_2m",
                        "openmeteo_dew_point_2m",
                        "openmeteo_precipitation",
                    ],
                    axis=1,
                )

                # Clean up rows with missing values
                self._weather_df = self._weather_df.dropna(subset=["air_temperature"])
                self._weather_df = self._weather_df.dropna(subset=["next_6h_symbol_code"])
                self._weather_df = self._weather_df.dropna(subset=["precipitation_amount"])

                self._weather_df["time_diff"] = self._weather_df["time"].diff()
                threshold = pd.Timedelta(hours=1)
                cutoff_index = self._weather_df[self._weather_df["time_diff"] > threshold].index.min()
                if pd.notna(cutoff_index):  # Check if there's a cutoff
                    self._weather_df = self._weather_df.loc[: cutoff_index - 1]
                self._weather_df = self._weather_df.drop(columns=["time_diff"])

                # Check if we got dew point temperatures. If not calculate them.
                has_nan_or_none = self._weather_df["dew_point_temperature"].isnull().any()
                if has_nan_or_none:
                    _LOGGER.warning(
                        f"Column 'dew_point_temperature' has NaN or None: {has_nan_or_none}. Calculating dew points."
                    )
                    self._weather_df["dew_point_temperature"] = await self._calculate_dew_point(self._weather_df)

                # Calculate atmosphere
                self._weather_df["seeing"] = await self._calculate_seeing(self._weather_df)
                self._weather_df["transparency"] = await self._calculate_transparency(self._weather_df)
                self._weather_df["lifted_index"] = await self._calculate_lifted_index(self._weather_df)

                self._test_weather_df()
            else:
                _LOGGER.debug("Using cached data")

    @typechecked
    def _get_location(
        self,
        latitude,
        longitude,
        elevation,
        timezone_info,
    ) -> GeoLocationData | None:
        """Returns a validated GeoLocation data object"""

        geolocation = GeoLocationDataModel(
            {
                "latitude": latitude,
                "longitude": longitude,
                "elevation": elevation,
                "timezone_info": timezone_info,
            }
        )

        try:
            return GeoLocationData(data=geolocation)
        except TypeError as ve:
            _LOGGER.error(f"Failed to parse geo location data: {geolocation}")
            _LOGGER.error(ve)
            return None

    @typechecked
    async def _get_condition(
        self,
        time,
    ) -> ConditionData | None:
        """Returns a validated Weather Conditions data object"""

        # Return row from dataframe
        timestamp = time.replace(tzinfo=None).strftime("%Y-%m-%d %H:00:00+00:00")
        row = self._weather_df.loc[self._weather_df["time"] == timestamp].iloc[0]
        
        cloudcover = float(row["cloud_area_fraction"])
        cloud_area_fraction = float(row["cloud_area_fraction"])
        cloudcover_high = float(row["cloud_area_fraction_high"])
        cloudcover_medium = float(row["cloud_area_fraction_medium"])
        cloudcover_low = float(row["cloud_area_fraction_low"])
        fog = float(row["fog_area_fraction"])
        rh2m = float(row["relative_humidity"])
        wind_speed = float(row["wind_speed"])
        wind_from_direction = float(row["wind_from_direction"])
        temp2m = float(row["air_temperature"])
        dewpoint2m = float(row["dew_point_temperature"])
        weather = str(row["next_1h_symbol_code"])
        weather6 = str(row["next_6h_symbol_code"])
        precipitation_amount = float(row["next_1h_precipitation_amount"])
        precipitation_amount6 = float(row["next_6h_precipitation_amount"])
        seeing = float(row["seeing"])
        transparency = float(row["transparency"])
        lifted_index = float(row["lifted_index"])

        # Calculate Fog Density
        fog2m = await self._atmosphere.calculate_fog_density(temp2m, rh2m, dewpoint2m, wind_speed) * 100

        condition = ConditionDataModel(
            {
                "cloudcover": cloudcover,
                "cloud_area_fraction": cloud_area_fraction,
                "cloud_area_fraction_high": cloudcover_high,
                "cloud_area_fraction_low": cloudcover_low,
                "cloud_area_fraction_medium": cloudcover_medium,
                "fog_area_fraction": max(fog, fog2m),
                "fog2m": fog2m,
                "seeing": seeing,
                "transparency": transparency,
                "lifted_index": lifted_index,
                "condition_percentage": await self._calc_condition_percentage(
                    cloudcover_high=cloudcover_high,
                    cloudcover_medium=cloudcover_medium,
                    cloudcover_low=cloudcover_low,
                    fog=fog,
                    fog2m=fog2m,
                    seeing=seeing,
                    transparency=transparency,
                    wind_speed=wind_speed,
                    precipitation_amount=precipitation_amount,
                ),
                "rh2m": rh2m,
                "wind_speed": wind_speed,
                "wind_from_direction": wind_from_direction,
                "temp2m": temp2m,
                "dewpoint2m": dewpoint2m,
                "weather": weather,
                "weather6": weather6,
                "precipitation_amount": precipitation_amount,
                "precipitation_amount6": precipitation_amount6,
            }
        )

        try:
            return ConditionData(data=condition)
        except TypeError as ve:
            _LOGGER.error(f"Failed to parse condition data: {condition}")
            _LOGGER.error(ve)
            return None

    @typechecked
    async def _calc_condition_percentage(
        self,
        cloudcover_high,
        cloudcover_medium,
        cloudcover_low,
        fog,
        fog2m,
        seeing,
        transparency,
        wind_speed,
        precipitation_amount,
    ) -> int:
        """Return condition based on cloud cover, fog, seeing, transparency, wind speed, and precipitation."""

        # Seeing is something in between 0 and 2.5 arcsecs
        seeing = seeing * 100 / SEEING_MAX  # arcsecs up to 2.5
        # transparency = int(transparency * 40)  # mag degration up to 2.5
        transparency = transparency * 100 / MAG_DEGRATION_MAX  # mag degration up to MAG_DEGRATION_MAX
        # Wind speed is something in between 0 and 16.5 m/s
        if wind_speed > WIND10M_MAX:
            wind_speed = WIND10M_MAX
        wind_speed_value = int(wind_speed * (100 / WIND10M_MAX))  # m/s up to 16.5

        cloudcover = []
        cloudcover.append(cloudcover_high * self._cloudcover_high_weakening)
        cloudcover.append(cloudcover_medium * self._cloudcover_medium_weakening)
        cloudcover.append(cloudcover_low * self._cloudcover_low_weakening)

        condition = int(
            100
            - (
                self._cloudcover_weight * max(cloudcover)
                + self._fog_weight * max(fog, fog2m)  # Use whatever is higher
                + self._seeing_weight * seeing
                + self._transparency_weight * transparency
                + self._calm_weight * wind_speed_value
            )
            / (
                self._cloudcover_weight
                + self._fog_weight
                + self._seeing_weight
                + self._transparency_weight
                + self._calm_weight
            )
            - precipitation_amount * 100
        )

        # _LOGGER.debug(
        #     f"Cloudcover: {max(cloudcover)}, Fog: {fog}, Seeing: {seeing}, Transparency: {transparency}, Wind speed: {wind_speed_value}"
        # )

        # Ensure condition is within the valid range [0, 1]
        condition = max(0, min(100, condition))

        return int(condition)

    @typechecked
    def _test_data(self, data, keys) -> bool:
        """Test that specific values in a dictionary are not None"""

        if not all(data[key] is not None for key in keys):
            return False
        if not all(data[key] != "NaN" for key in keys):
            return False
        return True

    def _test_weather_df(self) -> None:
        # Check main dataframe for any NaN values
        nan_locations = self._weather_df.isna()

        # Find column names and index positions of NaNs
        result = [
            (index, col)
            for col in nan_locations.columns
            for index in nan_locations.index
            if nan_locations.at[index, col]
        ]

        # Print results
        if result:
            _LOGGER.warning("NaN in main dataframe found at:")
            for index, col in result:
                _LOGGER.warning(f"Column: {col}, Index: {index}")
        else:
            _LOGGER.debug("No NaN values found in main dataframe.")

    # #########################################################################
    # Data for AstroWeather
    # #########################################################################
    @typechecked
    async def _get_location_data(self) -> List[LocationData]:
        """Returns a validated LocationData data object"""

        items = []

        await self._retrive_data()

        now = datetime.now(UTC).replace(tzinfo=None)

        if self._test_datetime is not None:
            await self._astro_routines.need_update()
        else:
            await self._astro_routines.need_update(forecast_time=now)

        forecast_time = now.replace(minute=0, second=0, microsecond=0)
        if self._test_datetime is not None:
            forecast_time = self._test_datetime.replace(minute=0, second=0, microsecond=0)
        _LOGGER.debug("Forecast time: %s", str(forecast_time))

        utc_to_local_diff = self._astro_routines.utc_to_local_diff()
        _LOGGER.debug("UTC to local diff: %s", str(utc_to_local_diff))
        
        if len(self._weather_df) == 0:
            _LOGGER.error("Weather data not available")
            return []

        data_index = int((self._weather_df["time"] == now.strftime("%Y-%m-%d %H:00:00+00:00")).idxmax())
        _LOGGER.debug("Data index: %s", str(data_index))

        time_data = TimeDataModel(
            {
                "forecast_time": forecast_time.replace(microsecond=0, tzinfo=timezone.utc),
            }
        )

        try:
            time_data = TimeData(data=time_data)
        except TypeError as ve:
            _LOGGER.error(f"Failed to parse location data model data: {time_data}")
            _LOGGER.error(ve)

        item = LocationDataModel(
            {
                # Time data
                "time_data": time_data,
                # Time shift to UTC
                "time_shift": await self._astro_routines.time_shift(),
                # Remaining forecast data point in met.no data
                "forecast_length": (len(self._weather_df) - data_index),
                # Location
                "location_data": self._location_data,
                # Astronomical routines
                "sun_data": await self._astro_routines.sun_data(),
                "moon_data": await self._astro_routines.moon_data(),
                "darkness_data": await self._astro_routines.darkness_data(),
                "night_duration_astronomical": await self._astro_routines.night_duration_astronomical(),
                "deepsky_forecast": await self._get_deepsky_forecast(),
                "condition_data": await self._get_condition(now),
                # Uptonight objects
                "uptonight": await self._get_deepsky_objects(),
                "uptonight_bodies": await self._get_bodies(),
                "uptonight_comets": await self._get_comets(),
            }
        )

        try:
            items.append(LocationData(data=item))
        except TypeError as ve:
            _LOGGER.error(f"Failed to parse location data model data: {item}")
            _LOGGER.error(ve)

        return items

    @typechecked
    async def _get_forecast_data(self, forecast_type, hours_to_show) -> List[ForecastData]:
        """Return Forecast data for the Station."""

        items = []

        await self._retrive_data()

        now = datetime.now(UTC).replace(tzinfo=None)

        # Create items
        cnt = 0

        forecast_time = now.replace(minute=0, second=0, microsecond=0)
        if self._test_datetime is not None:
            forecast_time = self._test_datetime.replace(minute=0, second=0, microsecond=0)
        _LOGGER.debug("Forecast time: %s", str(forecast_time))
        
        utc_to_local_diff = self._astro_routines.utc_to_local_diff()
        _LOGGER.debug("UTC to local diff: %s", str(utc_to_local_diff))

        if len(self._weather_df) == 0:
            _LOGGER.error("Weather data not available")
            return []

        data_index = int((self._weather_df["time"] == now.strftime("%Y-%m-%d %H:00:00+00:00")).idxmax())
        _LOGGER.debug("Data index: %s", str(data_index))

        for index, row in self._weather_df.iterrows():
            forecast_time = row["time"].replace(microsecond=0, tzinfo=timezone.utc)

            td = TimeDataModel(
                {
                    "forecast_time": forecast_time,  # timestamp
                }
            )

            try:
                time_data = TimeData(data=td)
            except TypeError as ve:
                _LOGGER.error(f"Failed to parse location data model data: {time_data}")
                _LOGGER.error(ve)

            item = None
            try:
                item = ForecastDataModel(
                    {
                        "time_data": time_data,  # Time data
                        "hour": row["time"].hour,  # forecast_time.hour % 24,
                        "condition_data": await self._get_condition(row["time"]),
                    }
                )
            except KeyError as ke:
                _LOGGER.error(f"Failed to parse location data model data: {ke}")
                _LOGGER.error(row["time"])
                
            try:
                if item is not None:
                    items.append(ForecastData(data=item))
            except TypeError as ve:
                _LOGGER.error(f"Failed to parse forecast data: {item}")
                _LOGGER.error(ve)

            cnt += 1
            if cnt >= hours_to_show:
                break

        self._forecast_data = items

        _LOGGER.debug("Forceast Length: %s", str(len(items)))

        return items

    @typechecked
    async def _get_deepsky_forecast(self) -> List[NightlyConditionsData]:
        """Return Deepsky Forecast data."""

        items = []

        if self._forecast_data is None:
            await self._get_forecast_data(FORECAST_TYPE_HOURLY, 72)

        utc_to_local_diff = self._astro_routines.utc_to_local_diff()
        _LOGGER.debug("UTC to local diff: %s", str(utc_to_local_diff))

        # Create forecast
        forecast_dayname = ""
        start_forecast_hour = 0
        start_weather = ""
        interval_points = []
        now = datetime.now(UTC).replace(tzinfo=None)

        if self._test_datetime is not None:
            await self._astro_routines.need_update()
        else:
            await self._astro_routines.need_update(forecast_time=now)

        sun_next_setting = await self._astro_routines.sun_next_setting()
        sun_next_rising = await self._astro_routines.sun_next_rising()
        night_duration_astronomical = await self._astro_routines.night_duration_astronomical()

        start_indexes = []
        # Find start index for two nights and store the indexes
        for index, details_forecast in enumerate(self._forecast_data):
            if details_forecast.forecast_time.hour % 24 == sun_next_rising.hour and len(start_indexes) == 0:
                start_indexes.append(0)
            if details_forecast.forecast_time.hour % 24 == sun_next_setting.hour:
                start_indexes.append(index)

        forecast_data_len = len(self._forecast_data)
        for day in range(0, len(start_indexes)):
            start_forecast_hour = 0
            start_weather = ""
            interval_points = []
            start_index = start_indexes[day]
            for index in range(
                start_index,
                start_index + int(math.floor(night_duration_astronomical / 3600) + 2),
            ):
                if index >= forecast_data_len:
                    _LOGGER.debug("No more forecast data")
                    break
                details_forecast = self._forecast_data[index]

                if len(interval_points) == 0:
                    forecast_dayname = details_forecast.forecast_time.strftime("%A")
                    start_forecast_hour = details_forecast.forecast_time.hour
                    start_weather = details_forecast.weather6
                    start_precipitation_amount6 = details_forecast.precipitation_amount6

                # Calculate Condition
                if len(interval_points) <= int(math.floor(night_duration_astronomical / 3600)):
                    interval_points.append(
                        await self._calc_condition_percentage(
                            cloudcover_high=details_forecast.cloud_area_fraction_high_percentage,
                            cloudcover_medium=details_forecast.cloud_area_fraction_medium_percentage,
                            cloudcover_low=details_forecast.cloud_area_fraction_low_percentage,
                            fog=details_forecast.fog_area_fraction_percentage,
                            fog2m=details_forecast.fog2m_area_fraction_percentage,
                            seeing=details_forecast.seeing,
                            transparency=details_forecast.transparency,
                            wind_speed=details_forecast.wind10m_speed,
                            precipitation_amount=details_forecast.precipitation_amount,
                        )
                    )

                if details_forecast.forecast_time.hour == sun_next_rising.hour or index >= (forecast_data_len - 1):
                    item = NightlyConditionsDataModel(
                        {
                            "dayname": forecast_dayname,
                            "hour": start_forecast_hour,
                            "nightly_conditions": interval_points,
                            "weather": start_weather,
                            "precipitation_amount6": start_precipitation_amount6,
                        }
                    )

                    try:
                        item = NightlyConditionsData(data=item)
                    except TypeError as ve:
                        _LOGGER.error(f"Failed to parse nightly conditions model data: {item}")
                        _LOGGER.error(ve)

                    items.append(item)

                    conditions_numeric = ""
                    for condition in interval_points:
                        conditions_numeric += str(condition) + ", "
                    _LOGGER.debug(
                        "Nightly conditions day: %s, start hour: %s, nightly conditions: %s, weather: %s, conditions numeric: %s",
                        str(forecast_dayname),
                        str(start_forecast_hour),
                        str(len(interval_points)),
                        str(start_weather),
                        conditions_numeric,
                    )

                    # Test for end of astronomical night. Will get true if we're already at night.
                    if details_forecast.forecast_time.hour % 24 == sun_next_rising.hour:
                        break
            start_index += 24

        return items

    # #########################################################################
    # UpTonight
    # #########################################################################
    @typechecked
    async def _retrieve_data_uptonight(self) -> None:
        """Retrieves current data from uptonight"""

        _LOGGER.debug("Updating data from uptonight")

        dataseries_dso = None
        dataseries_bodies = None
        dataseries_comets = None

        if os.path.exists(self._uptonight_path):
            if os.path.isfile(self._uptonight_path + "/uptonight-report.json"):
                # _LOGGER.debug(f"Uptonight report found")
                async with aiofiles.open(self._uptonight_path + "/uptonight-report.json", mode="r") as json_file:
                    contents = await json_file.read()
                dataseries_dso = json.loads(contents)
                _LOGGER.debug("Uptonight DSO imported")
            else:
                _LOGGER.debug(f"File uptonight-report.json not found in {self._uptonight_path}")

            if os.path.isfile(self._uptonight_path + "/uptonight-bodies-report.json"):
                # _LOGGER.debug(f"Uptonight report found")
                async with aiofiles.open(self._uptonight_path + "/uptonight-bodies-report.json", mode="r") as json_file:
                    contents = await json_file.read()
                dataseries_bodies = json.loads(contents)
                _LOGGER.debug("Uptonight Bodies imported")
            else:
                _LOGGER.debug(f"File uptonight-bodies-report.json not found in {self._uptonight_path}")

            if os.path.isfile(self._uptonight_path + "/uptonight-comets-report.json"):
                # _LOGGER.debug(f"Uptonight report found")
                async with aiofiles.open(self._uptonight_path + "/uptonight-comets-report.json", mode="r") as json_file:
                    contents = await json_file.read()
                dataseries_comets = json.loads(contents)
                _LOGGER.debug("Uptonight Comets imported")
            else:
                _LOGGER.debug(f"File uptonight-comets-report.json not found in {self._uptonight_path}")
        else:
            _LOGGER.debug(
                f"Path for UpTonight data not found. Current path: {self._uptonight_path}/uptonight-report.json"
            )

        self._weather_data_uptonight = dataseries_dso
        self._weather_data_uptonight_bodies = dataseries_bodies
        self._weather_data_uptonight_comets = dataseries_comets

    @typechecked
    async def _get_deepsky_objects(self) -> List[UpTonightDSOData]:
        """Return Deepsky Objects for today."""

        items = []

        # Create list of deep sky objects
        if self._weather_data_uptonight is not None:
            dso_id = self._weather_data_uptonight.get("id", {})
            dso_target_name = self._weather_data_uptonight.get("target name", {})
            dso_type = self._weather_data_uptonight.get("type", {})
            dso_constellation = self._weather_data_uptonight.get("constellation", {})
            dso_size = self._weather_data_uptonight.get("size", {})
            dso_visual_magnitude = self._weather_data_uptonight.get("mag", {})
            dso_meridian_transit = self._weather_data_uptonight.get("meridian transit", {})
            dso_meridian_antitransit = self._weather_data_uptonight.get("meridian antitransit", {})
            dso_foto = self._weather_data_uptonight.get("foto", {})

            for row in range(len(dso_target_name)):
                dso_meridian_transit_local = dso_meridian_transit.get(str(row), "")
                if dso_meridian_transit_local != "":
                    dso_meridian_transit_utc = (
                        datetime.strptime(dso_meridian_transit_local, "%m/%d/%Y %H:%M:%S")
                        - timedelta(seconds=await self._astro_routines.time_shift())
                    ).replace(tzinfo=UTC)
                else:
                    dso_meridian_transit_utc = ""

                dso_meridian_antitransit_local = dso_meridian_antitransit.get(str(row), "")
                if dso_meridian_antitransit_local != "":
                    dso_meridian_antitransit_utc = (
                        datetime.strptime(dso_meridian_antitransit_local, "%m/%d/%Y %H:%M:%S")
                        - timedelta(seconds=await self._astro_routines.time_shift())
                    ).replace(tzinfo=UTC)
                else:
                    dso_meridian_antitransit_utc = ""

                item = UpTonightDSODataModel(
                    {
                        "id": dso_id.get(str(row), ""),
                        "target_name": dso_target_name.get(str(row), ""),
                        "type": dso_type.get(str(row), ""),
                        "constellation": dso_constellation.get(str(row), ""),
                        "size": dso_size.get(str(row), 0),
                        "visual_magnitude": dso_visual_magnitude.get(str(row), 0),
                        "meridian_transit": dso_meridian_transit_utc,
                        "meridian_antitransit": dso_meridian_antitransit_utc,
                        "foto": dso_foto.get(str(row), 0),
                    }
                )
                try:
                    items.append(UpTonightDSOData(data=item))
                except TypeError as ve:
                    _LOGGER.error(f"Failed to parse deep sky object data: {item}")
                    _LOGGER.error(ve)

        return items

    @typechecked
    async def _get_bodies(self) -> List[UpTonightBodiesData]:
        """Return Bodies for today."""

        items = []

        # Create list of bodies
        if self._weather_data_uptonight_bodies is not None:
            body_target_name = self._weather_data_uptonight_bodies.get("target name", {})
            body_max_altitude = self._weather_data_uptonight_bodies.get("max altitude", {})
            body_azimuth = self._weather_data_uptonight_bodies.get("azimuth", {})
            body_max_altitude_time = self._weather_data_uptonight_bodies.get("max altitude time", {})
            body_visual_magnitude = self._weather_data_uptonight_bodies.get("visual magnitude", {})
            body_meridian_transit = self._weather_data_uptonight_bodies.get("meridian transit", {})
            body_foto = self._weather_data_uptonight_bodies.get("foto", {})

            for row in range(len(body_target_name)):
                # UpTonight delivers the time in local time zone. here we need it in UTC
                body_max_altitude_time_local = body_max_altitude_time.get(str(row), "")
                if body_max_altitude_time_local != "":
                    body_max_altitude_time_utc = (
                        datetime.strptime(body_max_altitude_time_local, "%m/%d/%Y %H:%M:%S")
                        - timedelta(seconds=await self._astro_routines.time_shift())
                    ).replace(tzinfo=UTC)
                else:
                    body_max_altitude_time_utc = ""

                body_meridian_transit_local = body_meridian_transit.get(str(row), "")
                if body_meridian_transit_local != "":
                    body_meridian_transit_utc = (
                        datetime.strptime(body_meridian_transit_local, "%m/%d/%Y %H:%M:%S")
                        - timedelta(seconds=await self._astro_routines.time_shift())
                    ).replace(tzinfo=UTC)
                else:
                    body_meridian_transit_utc = ""

                item = UpTonightBodiesDataModel(
                    {
                        "target_name": body_target_name.get(str(row), ""),
                        "max_altitude": body_max_altitude.get(str(row), 0),
                        "azimuth": body_azimuth.get(str(row), 0),
                        "max_altitude_time": body_max_altitude_time_utc,
                        "visual_magnitude": body_visual_magnitude.get(str(row), 0),
                        "meridian_transit": body_meridian_transit_utc,
                        "foto": body_foto.get(str(row), 0),
                    }
                )
                try:
                    items.append(UpTonightBodiesData(data=item))
                except TypeError as ve:
                    _LOGGER.error(f"Failed to parse bodies data: {item}")
                    _LOGGER.error(ve)

        return items

    @typechecked
    async def _get_comets(self) -> List[UpTonightCometsData]:
        """Return Comets for today."""

        items = []

        # Create list of comets
        if self._weather_data_uptonight_comets is not None:
            comet_target_name = self._weather_data_uptonight_comets.get("target name", {})
            distance_au_earth = self._weather_data_uptonight_comets.get("distance earth au", {})
            distance_au_sun = self._weather_data_uptonight_comets.get("distance sun au", {})
            absolute_magnitude = self._weather_data_uptonight_comets.get("absolute magnitude", {})
            visual_magnitude = self._weather_data_uptonight_comets.get("visual magnitude", {})
            altitude = self._weather_data_uptonight_comets.get("altitude", {})
            azimuth = self._weather_data_uptonight_comets.get("azimuth", {})
            rise_time = self._weather_data_uptonight_comets.get("rise time", {})
            set_time = self._weather_data_uptonight_comets.get("set time", {})

            for row in range(len(comet_target_name)):
                # UpTonight delivers the time in local time zone. here we need it in UTC
                rise_time_local = rise_time.get(str(row), "")
                if rise_time_local != "":
                    rise_time_local_utc = (
                        datetime.strptime(rise_time_local, "%m/%d/%Y %H:%M:%S")
                        - timedelta(seconds=await self._astro_routines.time_shift())
                    ).replace(tzinfo=UTC)
                else:
                    rise_time_local_utc = ""

                set_time_local = set_time.get(str(row), "")
                if set_time_local != "":
                    set_time_local_utc = (
                        datetime.strptime(set_time_local, "%m/%d/%Y %H:%M:%S")
                        - timedelta(seconds=await self._astro_routines.time_shift())
                    ).replace(tzinfo=UTC)
                else:
                    set_time_local_utc = ""

                item = UpTonightCometsDataModel(
                    {
                        "designation": comet_target_name.get(str(row), ""),
                        "distance_au_earth": distance_au_earth.get(str(row), 0),
                        "distance_au_sun": distance_au_sun.get(str(row), 0),
                        "absolute_magnitude": absolute_magnitude.get(str(row), 0),
                        "visual_magnitude": visual_magnitude.get(str(row), 0),
                        "altitude": altitude.get(str(row), 0),
                        "azimuth": azimuth.get(str(row), 0),
                        "rise_time": rise_time_local_utc,
                        "set_time": set_time_local_utc,
                    }
                )
                try:
                    items.append(UpTonightCometsData(data=item))
                except TypeError as ve:
                    _LOGGER.error(f"Failed to parse comets data: {item}")
                    _LOGGER.error(ve)

        return items

    # #########################################################################
    # Met.no
    # #########################################################################
    async def _retrieve_data_metno(self) -> None:
        """Retrieves current data from met."""

        _LOGGER.debug("Updating data from Met.no")

        dataseries = None

        # Testing
        if self._test_mode:
            if os.path.isfile("debug/met.json"):
                _LOGGER.debug("Reading Met.no data from file")
                with open("debug/met.json") as json_file:
                    dataseries = json.load(json_file).get("properties", {}).get("timeseries")
            else:
                json_data_metno = await self._async_request_met()
                dataseries = json_data_metno.get("properties", {}).get("timeseries")
        else:
            json_data_metno = await self._async_request_met()
            dataseries = json_data_metno.get("properties", {}).get("timeseries")

        # Flattening the data into a list of records
        records = []
        if dataseries is not None:
            for entry in dataseries:
                time = datetime.strptime(entry["time"], "%Y-%m-%dT%H:%M:%SZ").replace(
                    microsecond=0, tzinfo=timezone.utc
                )
                instant_details = entry["data"]["instant"]["details"]
                next_1h_symbol_code = entry["data"].get("next_1_hours", {}).get("summary", {}).get("symbol_code", None)
                next_6h_symbol_code = entry["data"].get("next_6_hours", {}).get("summary", {}).get("symbol_code", None)
                next_1h_precipitation_amount = (
                    entry["data"].get("next_1_hours", {}).get("details", {}).get("precipitation_amount", None)
                )
                next_6h_precipitation_amount = (
                    entry["data"].get("next_6_hours", {}).get("details", {}).get("precipitation_amount", None)
                )

                # Adding all details to a single record
                record = {
                    "time": time,
                    **instant_details,
                    "next_1h_symbol_code": next_1h_symbol_code,
                    "next_6h_symbol_code": next_6h_symbol_code,
                    "next_1h_precipitation_amount": next_1h_precipitation_amount,
                    "next_6h_precipitation_amount": next_6h_precipitation_amount,
                }
                records.append(record)

            # Convert to DataFrame
            weather_df_metno = pd.DataFrame(records)

            return weather_df_metno
        else:
            return None

    @typechecked
    async def _async_request_met(self) -> Dict:
        """Make a request against the Met.no API."""

        use_running_session = self._session and not self._session.closed

        if use_running_session:
            session = self._session
        else:
            session = ClientSession(
                timeout=ClientTimeout(total=DEFAULT_TIMEOUT),
            )

        # BASE_URL_MET = "https://api.met.no/weatherapi/locationforecast/2.0/complete?altitude=XX&lat=XX.XX&lon=XX.XX"
        url = (
            str(f"{BASE_URL_MET}")
            + "?lon="
            + str("%.1f" % round(self._location_data.longitude, 2))
            + "&lat="
            + str("%.1f" % round(self._location_data.latitude, 2))
            + "&altitude="
            + str(int(self._location_data.elevation))
        )

        try:
            _LOGGER.debug(f"Query url: {url}")
            async with session.request("get", url, headers=HEADERS) as resp:
                resp.raise_for_status()
                # plain = str(await resp.text()).replace("\n", " ")
                # data = json.loads(plain)
                data = await resp.json()

                if self._test_mode:
                    json_string = json.dumps(data)
                    with open("debug/met.json", "w") as outfile:
                        outfile.write(json_string)

                return data
        except JSONDecodeError as jsonerr:
            _LOGGER.error(f"JSON decode error, expecting value: {jsonerr}")
            return {}
        except asyncio.TimeoutError as tex:
            _LOGGER.error(f"Request to endpoint timed out: {tex}")
            return {}
        except ClientError as err:
            _LOGGER.error(f"Error requesting data: {err}")
            return {}

        finally:
            if not use_running_session:
                await session.close()

    # #########################################################################
    # Openmeteo
    # #########################################################################
    async def _retrieve_data_openmeteo(self) -> None:
        """Retrieves current data from openmeteo."""

        _LOGGER.debug("Updating data from Open-Meteo")

        use_running_session = self._session and not self._session.closed

        if use_running_session:
            session = self._session
        else:
            session = ClientSession(timeout=ClientTimeout(total=DEFAULT_TIMEOUT))

        params = {
            "latitude": self._location_data.latitude,
            "longitude": self._location_data.longitude,
            "hourly": [
                "temperature_2m",
                "relative_humidity_2m",
                "dew_point_2m",
                "precipitation",
                "cloud_cover",
                "cloud_cover_low",
                "cloud_cover_mid",
                "cloud_cover_high",
                "wind_speed_10m",
                "wind_direction_10m",
            ],
            "timezone": "UTC",
            # "temperature_unit": "C",
            "precipitation_unit": "mm",
            "wind_speed_unit": "ms",
            "models": self._forecast_model,
        }

        try:
            _LOGGER.debug(f"Query url: {BASE_URL_OPENMETEO}")

            # async with asyncio.timeout(self.request_timeout):
            response = await session.get(BASE_URL_OPENMETEO, params=params)
        except asyncio.TimeoutError as exception:
            msg = "Timeout occurred while connecting to the Open-Meteo API"
            raise OpenMeteoConnectionError(msg) from exception
        except (
            ClientError,
            ClientResponseError,
            socket.gaierror,
        ) as exception:
            msg = "Error occurred while communicating with Open-Meteo API"
            raise OpenMeteoConnectionError(msg) from exception
        finally:
            if not use_running_session:
                await session.close()

        content_type = response.headers.get("Content-Type", "")

        # Did we get a 4xx or 5xx?
        if (response.status // 100) in [4, 5]:
            if "application/json" in content_type:
                data = await response.json()
                response.close()
                if data.get("error") is True and (reason := data.get("reason")):
                    raise OpenMeteoError(reason)
                raise OpenMeteoError(response.status, data)
            contents = await response.read()
            response.close()
            raise OpenMeteoError(response.status, {"message": contents.decode("utf8")})

        text = await response.text()
        if "application/json" not in content_type:
            msg = "Unexpected response from the Open-Meteo API"
            raise OpenMeteoError(
                msg,
                {"Content-Type": content_type, "response": text},
            )

        # Process hourly data
        data = await response.json()
        if self._test_mode:
            json_string = json.dumps(data)
            with open("debug/openmeteo.json", "w") as outfile:
                outfile.write(json_string)

        response.close()
        hourly = data.get("hourly")

        hourly_data = {
            "time": np.array(hourly.get("time")),
            "openmeteo_temperature_2m": np.array(hourly.get("temperature_2m")),
            "openmeteo_relative_humidity_2m": np.array(hourly.get("relative_humidity_2m")),
            "openmeteo_dew_point_2m": np.array(hourly.get("dew_point_2m")),
            "openmeteo_precipitation": np.array(hourly.get("precipitation")),
            "openmeteo_cloud_cover": np.array(hourly.get("cloud_cover")),
            "openmeteo_cloud_cover_low": np.array(hourly.get("cloud_cover_low")),
            "openmeteo_cloud_cover_mid": np.array(hourly.get("cloud_cover_mid")),
            "openmeteo_cloud_cover_high": np.array(hourly.get("cloud_cover_high")),
            "openmeteo_wind_speed_10m": np.array(hourly.get("wind_speed_10m")),
            "openmeteo_wind_direction_10m": np.array(hourly.get("wind_direction_10m")),
        }

        weather_df_openmeteo = pd.DataFrame(data=hourly_data)
        weather_df_openmeteo["time"] = pd.to_datetime(weather_df_openmeteo["time"])
        weather_df_openmeteo["time"] = weather_df_openmeteo["time"].dt.tz_localize("UTC")

        return weather_df_openmeteo



"""Constant Definitions for AstroWeather."""

# #####################################################
# Requests
# #####################################################
BASE_URL_OPENMETEO = "https://api.open-meteo.com/v1/forecast"
BASE_URL_MET = "https://api.met.no/weatherapi/locationforecast/2.0/complete"
HEADERS = {"User-Agent": "AstroWeather github.com/mawinkler/astroweather"}

# #####################################################
# Defaults
# #####################################################
DEFAULT_TIMEOUT = 10
DEFAULT_CACHE_TIMEOUT = 1770
DEFAULT_TIMEZONE = "Etc/UTC"
DEFAULT_LATITUDE = 0.0
DEFAULT_LONGITUDE = 0.0
DEFAULT_ELEVATION = 0

# #####################################################
# Twilight and dusk
# #####################################################
CIVIL_TWILIGHT = 0
CIVIL_DUSK_DAWN = -6
NAUTICAL_TWILIGHT = -6
NAUTICAL_DUSK_DAWN = -12
ASTRONOMICAL_TWILIGHT = -12
ASTRONOMICAL_DUSK_DAWN = -18

# #####################################################
# Cloudcover
# #####################################################
# CLOUDCOVER_PLAIN = [
#     "0 to 6%",
#     "6 to 19%",
#     "19 to 31%",
#     "31 to 44%",
#     "44 to 56%",
#     "56 to 69%",
#     "69 to 81%",
#     "81 to 94%",
#     "94 to 100%",
# ]

# #####################################################
# Seeing
# #####################################################
SEEING = [0.25, 0.625, 0.875, 1.125, 1.375, 1.75, 2.25, 2.5]
# SEEING_PLAIN = [
#     'Below 0.5"',
#     '0.5 to 0.75"',
#     '0.75 to 1"',
#     '1 to 1.25"',
#     '1.25 to 1.5"',
#     '1.5 to 2"',
#     '2 to 2.5"',
#     'Over 2.5"',
# ]
SEEING_MAX = 2.5

# #####################################################
# Transparency
# #####################################################
TRANSPARENCY = [0.15, 0.35, 0.45, 0.55, 0.65, 0.775, 0.925, 1]
MAG_DEGRATION_MAX = 1
# TRANSPARENCY_PLAIN = [
#     "Below 0.3 mag",
#     "0.3 to 0.4 mag",
#     "0.4 to 0.5 mag",
#     "0.5 to 0.6 mag",
#     "0.6 to 0.7 mag",
#     "0.7 to 0.85 mag",
#     "0.85 to 1 mag",
#     "Over 1 mag",
# ]

# #####################################################
# Lifted Index
# #####################################################
# Negative Values (Unstable Atmosphere):
# LI < -6: Strongly unstable atmosphere, potential for severe thunderstorms and severe weather outbreaks.
# LI ≈ -4 to -6: Moderately unstable atmosphere, potential for thunderstorms and severe weather.
# LI ≈ -2 to -4: Slightly unstable atmosphere, some potential for thunderstorms.
# Near Zero Values:
# LI ≈ 0: Neutral stability, some potential for thunderstorms, weak to moderate vertical motion.
# Positive Values (Stable Atmosphere):
# LI ≈ +1 to +3: Slightly stable atmosphere, limited potential for thunderstorms, weak vertical motion.
# LI ≈ +4 to +6: Moderately stable atmosphere, very limited potential for thunderstorms, very weak vertical motion.
# LI > +6: Strongly stable atmosphere, very limited potential for thunderstorms, very weak to no vertical motion.
LIFTED_INDEX_VALUE = [1, 2, 3, 4, 5, 6, 7, 8]
# LIFTED_INDEX_RANGE = [
#     (-20, -6.99),
#     (-7, -4.99),
#     (-5, -2.99),
#     (-3, -0.01),
#     (0, 3.99),
#     (4, 7.99),
#     (8, 10.99),
#     (11, 20),
# ]
LIFTED_INDEX_RANGE = [
    (-7, -6.01),
    (-6, -4.01),
    (-4, -1.51),
    (-1.5, -0.01),
    (0, 1.49),
    (1.5, 3.99),
    (4, 5.99),
    (6, 7),
]
LIFTED_INDEX_PLAIN = [
    "Below -6, very unstable",
    "-6 to -4, very unstable",
    "-4 to -1.5, unstable",
    "-1.5 to 0, slightly unstable",
    "0 to 1.5, stable",
    "1.5 to 4, stable",
    "4 to 6, very stable",
    "Over 6, very stable",
]

# #####################################################
# Wind
# #####################################################
# 0 --- Calm            less than 1 mph (0 m/s)      Smoke rises vertically
# 1 --- Light air       1 - 3 mph     0.5-1.5 m/s    Smoke drifts with air, weather vanes inactive
# 2 --- Light breeze    4 - 7 mph     2-3 m/s        Weather vanes active, wind felt on face, leaves rustle
# 3 --- Gentle breeze   8 - 12 mph    3.5-5 m/s      Leaves & small twigs move, light flags extend
# 4 --- Moderate breeze 13 - 18 mph   5.5-8 m/s      Small branches sway, dust & loose paper blows about
# 5 --- Fresh breeze    19 - 24 mph   8.5-10.5 m/s   Small trees sway, waves break on inland waters
# 6 --- Strong breeze   25 - 31 mph   11-13.5 m/s    Large branches sway, umbrellas difficult to use
# 7 --- Moderate gale   32 - 38 mph   14-16.5 m/s    Whole trees sway, difficult to walk against wind
# Too windy
# 8 --- Fresh gale      39 - 46 mph   17-20 m/s      Twigs broken off trees, walking against wind very difficult
# 9 --- Strong gale     47 - 54 mph   20.5-23.5 m/s  Slight damage to buildings, shingles blown off roof
# 10 -- Whole gale      55 - 63 mph   24-27.5 m/s    Trees uprooted, considerable damage to buildings
# 11 -- Storm           64 - 73 mph   28-31.5 m/s    Widespread damage, very rare occurrence
# 12 -- Hurricane       over 73 mph   over 32 m/s    Violent destruction
WIND10M_DIRECTON = ["N", "NE", "E", "SE", "S", "SW", "W", "NW"]
WIND10M_VALUE = [1, 2, 3, 4, 5, 6, 7, 8]
WIND10M_RANGE = [
    (0, 0.49),
    (0.5, 1.99),
    (2, 3.49),
    (3.5, 5.49),
    (5.5, 8.49),
    (8.5, 10.99),
    (11, 13.99),
    (14, 100),
]
WIND10M_PLAIN = [
    "Calm",
    "Light air",
    "Light breeze",
    "Gentle breeze",
    "Moderate breeze",
    "Fresh breeze",
    "Strong breeze",
    "Moderate gale",
]
WIND10M_MAX = 16.5

# #####################################################
# Condition
# #####################################################
DEFAULT_CONDITION_CLOUDCOVER_WEIGHT = 3
DEFAULT_CONDITION_CLOUDCOVER_HIGH_WEAKENING = 1
DEFAULT_CONDITION_CLOUDCOVER_MEDIUM_WEAKENING = 1
DEFAULT_CONDITION_CLOUDCOVER_LOW_WEAKENING = 1
DEFAULT_CONDITION_FOG_WEIGHT = 3
DEFAULT_CONDITION_SEEING_WEIGHT = 2
DEFAULT_CONDITION_TRANSPARENCY_WEIGHT = 1
DEFAULT_CONDITION_CALM_WEIGHT = 2
CONDITION_PLAIN = ["excellent", "good", "fair", "poor", "bad"]
CONDITION = ["█", "▆", "▄", "▂", "▁"]
DEEP_SKY_THRESHOLD = 75
DARK_NIGHT_MAX_MOON_PHASE = 0.05
DARK_NIGHT_MAX_MOON_ALT = "5:0:0"

# #####################################################
# Forecast
# #####################################################
FORECAST_TYPE_DAILY = "daily"
FORECAST_TYPE_HOURLY = "hourly"
FORECAST_TYPES = [
    FORECAST_TYPE_DAILY,
    FORECAST_TYPE_HOURLY,
]
